# pygtktest.py
# 
# Very short test file to see if the installation of PyGTK was successful.
#
# Author       : Chris X. Weichenberger
# Created      : unk
# License
# Twilight is a GUI for displaying ligands incompatible with electron density.
#    Copyright (C) 2012  Christian X. Weichenberger, Edwin Pozharski, and
#    Bernhard Rupp.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import pygtk
pygtk.require('2.0')
import gtk
import gobject;

window = gtk.MessageDialog(buttons=gtk.BUTTONS_OK_CANCEL);
window.connect("destroy", lambda wid: gtk.main_quit());
window.connect("delete_event", lambda a1,a2:gtk.main_quit());
window.set_property("text", "You did it!")
window.run()



# optconf.py
# 
# User preferences for dysplayer.
#
# Author       : Ed Pozharski
# Co-author    : Chris X. Weichenberger
# Created      : unk
# License
# Twilight is a GUI for displaying ligands incompatible with electron density.
#    Copyright (C) 2012  Christian X. Weichenberger, Edwin Pozharski, and
#    Bernhard Rupp.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Modifications:
#  2012-02-22 (_X) Adaption to dysplayer with default values and handling.
#  2012-03-14 (_X) Added scripting host.
#  2012-03-27 (_X) Changed default name of linux config file to .dysplayerrc
#    and made it platform specifig, under Win$ it is dysplayer.cfg
#  2012-05-09 (_X) Changed annotation text.
#  2012-08-01 (_X) Platform dependent default for coot executable.


""" Program configuration classes. General idea is of a class that has
    the dictionary of options as data and includes various methods to:
    a) create ConfigParser from scratch;
    b) Read/write configuration files;
    c) Update ConfigParser to the current version of the dictionary."""

import os;
import tempfile;
import ConfigParser;

from files import Output;
import sys
try:
    from collections import OrderedDict as ODict;
except ImportError:
    ODict = dict ;
    Output("Warning: Configuration file will contain shuffled output.");


# Platform-specific configuration file name.
DEFAULT_CFG_FN = "dysplayer.cfg" if( sys.platform=="win32" ) else ".dysplayerrc";

# Default sigma contour level for 2Fo - Fc maps.
SIGMA_2FOFC         = "1.0";
# Default sigma contour level for  Fo - Fc maps.
SIGMA_FOFC          = "3.0";
# Default width of coot window (in pixels).
COOT_WIDTH          = "600";
# Default height of coot window (in pixels).
COOT_HEIGHT         = "600";
# Default coot executable name, for Windows, this is win32, regardless of arch.
COOT_EXEC           = "runwincoot.bat" if( sys.platform=="win32" ) else "coot";
# Scripting engine for coot. Either 'python' or 'scheme'.
COOT_SCRIPTING_HOST = "python";

# We need the ordered dictionary here as we do want to preserve the order
# in the config file. This is of quite some importance, as the menu is
# constructed based on this order!
#
# Funny enough, the options are case INsensitive.
DEFAULT_CONFIGURATION = ODict(
    # program paramerters 
    (("program",
      ODict( (("2fofc", SIGMA_2FOFC),
              ("fofc", SIGMA_FOFC),
              ("cootexec", COOT_EXEC),
              ("pdbdir", ""),
              ("tempdir", tempfile.gettempdir()), 
              ("ustamp", ""),
              ("coot-width", COOT_WIDTH),
              ("coot-height", COOT_HEIGHT),
              ("scripting-host", COOT_SCRIPTING_HOST)
              ))
      ),
     ("annotation",
      ODict( (("1", "No evidence - Ligand forced into empty site."),
              ("2", "Projection - Partly disordered, placed by extension."), 
              ("3", "Misplacement - Misplaced ligand close to its density."),
              ("4", "Mother liquor - Ligand placed into buffer density."),
              ("5", "Petering out - Ligand density vanishes further out."),
              ("6", "False positive - False positive. Sorry!"),
              ))
      ),
     )
);


_Configurer = None;
_Config     = None;

class CConfigurer:
    def __init__(self, sections=None, cfgname=DEFAULT_CFG_FN,
                 defaultarget=None):
        """ Configurator class holds the options dictionary and provides the 
            methods needed to create/read/write/update the configuration
            parsers.  See ConfigParser module for more details on the way
            python handles configuration files.
            
            Class instance parameters are:
            
            sections     - dictionary of option sections, each value is a 
                            dictionary of options containing keys and default 
                            values.
            cfgname      - the configuration file name. 
            defaultarget - the default location of the config file.  By default,
                            is set to user's home directory.
        """            
        if sections:
            self.sections = sections
        else:
            self.sections = {}
        self.cfgname = cfgname
        self.SetDefaulTarget(defaultarget);

    def SetConfigFileName(self, cfgname=DEFAULT_CFG_FN):
        """ The configuration file name is normally given when the class is 
            invoked and maintained for the lifetime of the class instance.
            This method allows to change the file name if needed.  Use with
            caution since the next configuration reading event may fail if
            the file does not exist.  On the other hand, this method may be
            useful for switching between different versions of the configuration,
            but that may also be done by using two instances of the class."""
        self.cfgname = cfgname

    def SetDictionary(self, sections):
        """ The dictionary that defines the configuration file is normally
            initiated when a class instanec is created and should in most
            circumstances be maintained for the lifetime of the class instance.
            This method covers a presumably esoteric situation when the 
            items in the configuration file default have to be changed for an
            existing class instance.  Use with caution, it may be more sane to
            just create another instance. """
        self.sections = sections 

    def SetDefaulTarget(self, defaultarget=None):
        """ The default location of the configuration file should be generally
            provided upon class instantiation and kept unchanged.  This method
            may be invoked (use caution) to change the default location during
            for an existing class instance. Defaults to current user home
            folder."""
        if defaultarget is None:
            self.defaultarget = os.path.expanduser('~')
        else:
            self.defaultarget = defaultarget

    def zero_config(self, target=None):
        """ Initialize the configuration file and drop it in the target folder.
            Returns the ConfigParser object it created. """
        if not target:
            target = self.defaultarget
        config = ConfigParser.SafeConfigParser()
        for section in self.sections.keys():
            config.add_section(section)
            for option in self.sections[section].keys():
                config.set(section, option, self.sections[section][option])
        fconf=open(os.path.join(target,self.cfgname), 'wb')
        config.write(fconf)
        fconf.close()
        return config

    def read_config(self, target=None):
        """ Read and return the ConfigParser object with configuration 
            from target folder. """
        if not target:
            target = self.defaultarget
        config = ConfigParser.SafeConfigParser()
        config.read(os.path.join(target,self.cfgname))
        return config

    def write_config(self, config, target=None):
        """ Write the configuration file defined by the ConfigParser instance
            passed to the method into target folder. """
        if not target:
            target = self.defaultarget
        fconf=open(os.path.join(target,self.cfgname), 'wb')
        config.write(fconf)
        fconf.close()

    def adjust_config(self, target=None, fWrite=True,
                      fRemoveMissingSections=True,
                      fRemoveMissingOptions=False):
        """ Read the configuration from target folder and fill out the missing 
            fields. By default, the adjusted configuration is written in to the
            target folder, this may be prevented by setting fWrite to False.
            For reasosn I don't quite remember, individual options are not 
            removed by default if the updated dictionary is missing them, but 
            the whole sections are.  This behavior may be changed by setting
            fRemoveMissingSections/fRemoveMissingOptions which are quite
            self-explanatory.  
            The return value is the adjusted configuration parser."""
        if not target:
            target = self.defaultarget
        config = self.read_config(target)
        for section in self.sections.keys():
            if not config.has_section(section):
                config.add_section(section)
            for option in self.sections[section].keys():
                if not config.has_option(section, option):
                    config.set(section, option, self.sections[section][option])
        if fRemoveMissingSections:
            for section in config.sections():
                if not self.sections.has_key(section):
                    config.remove_section(section)
        if fRemoveMissingOptions:
            for section in config.sections():
                for option in config.options(section):
                    if not self.sections[section].has_key(option):
                        config.remove_option(section,option)
        if fWrite:
            self.write_config(config,target)
        return config

    def found_config(self, target=None, mode=os.F_OK):
        """ Check the status of the configuration file in the target folder. 
            The type of access check is defined by the mode parameter, which 
            defaults to file existence and follows the same rules as os.access()
            method mode parameter.  If the target folderr is not specified,
            check the default target of the class."""
        if not target:
            target = self.defaultarget
        return os.access(os.path.join(target,self.cfgname),mode)



def die(text, exitCode=1):
    """Convenience function for ungraceful exit."""
    Output(text);
    sys.exit(exitCode);

class CDysConfigurer(CConfigurer):
    """Special configuration class for Dysplayer.

    Maintains the config object itself and provides get/set functions for
    it. Generates a new config object, if there is no config file. Otherwise,
    it reads the config file. Makes some basic sanity checks on the specific
    dysplayer config file contents and quits the application if violated.
    """
    def __init__(s, sections=DEFAULT_CONFIGURATION, cfgname=DEFAULT_CFG_FN,
                 defaultarget=None):
        CConfigurer.__init__(s, sections=sections, cfgname=cfgname,
                             defaultarget=defaultarget);
        s.Config = s.adjust_config() if( s.found_config() ) else s.zero_config();
        
        # we might brutally exit the program in the next function call as an
        # error detected there is not revoverable.
        s.IntegrityCheck();
        
    def GetCurrentConfigFullFileName(self):
        """Get the config file full path name.
        """
        return os.path.join(self.defaultarget, self.cfgname);
        
    def IntegrityCheck(s):
        """Performs integrity check of the configuration file.
        
        If these checks fail, the config file is irreversibly messed up and
        needs to be re-edited by the user, who in fact has caused this mess...
        """
        # full path config file name, either default or user defined.
        cfgFFN = os.path.join(s.defaultarget, s.cfgname);
        splitText = cfgFFN;
        splitText+= ", Token %s: Please specify exactly a single split character '-'.";
        notEnoughText = cfgFFN;
        notEnoughText+= ", Token %s: Empty text for menu entry or table annotation.";
        for key, text in s.Config.items("annotation"):
            menuTexts = text.split("-");
            if( len(menuTexts)!=2 ):
                Output(splitText % key);
                sys.exit(1);
            menuText, tableText = menuTexts;
            menuText, tableText = menuText.strip(), tableText.strip();
            if( menuText=="" or tableText=="" ):
                die(notEnoughText % key);
        
        notFloatText= cfgFFN+", ``%s'' does not specify a float value.";
        try:
            for var in ("2FoFc", "FoFc"):
                dummy = float(s.Config.get("program", var));
        except ValueError:
            die(notFloatText % var);

        notIntText= cfgFFN +", ``%s'' does not specify an int value.";
        try:
            for var in ("coot-width", "coot-height"):
                dummy = int(s.Config.get("program", var));
        except ValueError:
            die(notIntText % var);

        notADirText = cfgFFN+", ``%s'' does not specify an existing directory.";
        notCdText = cfgFFN +", Cannot change to directory specified by ``%s''.";
        for var in ("pdbDir", "tempDir"):
            dirName = s.Config.get("program", var);
            if( dirName!="" and not os.path.isdir(dirName) ):
                die(notADirText % var);
            if( dirName!="" and not os.access(dirName, os.X_OK) ):
                die(notCdText % var);

        notWriteText=cfgFFN +", Cannot write to directory specified by ``%s''.";
        tempDirName = s.Config.get("program", "tempDir");
        if( tempDirName!="" and not os.access(tempDirName, os.W_OK) ):
            die(notWriteText % var);
        
        badScriptingHostText=cfgFFN +", Wrong scripting language, please use "
        "either 'python' or 'scheme'.";
        scriptLang = s.Config.get("program", "scripting-host");
        if( scriptLang not in ("python", "scheme") ):
            die(badScriptingHostText);
            
    def WriteConfiguration(s):
        """Saves the configuration as a SafeConfigParser file."""
        s.write_config(s.Config);
    
    def GetConfig(s):
        """Get the actual SafeConfigParser object."""
        return s.Config;



#!/usr/bin/env python
# 
# Displays a table and calls coot.
#
# Author       : Chris Weichenberger
# Created      : 2011-10-26
# License
# Twilight is a GUI for displaying ligands incompatible with electron density.
#    Copyright (C) 2012  Christian X. Weichenberger, Edwin Pozharski, and
#    Bernhard Rupp.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Modifications:
# 2011-12-18 (CW) First release to the public. (v1.0)
# 2011-12-22 (EP) Added coot executable. (v1.1)
# 2012-02-08 (CW) Added sigma contour level as a parameter. (v1.2)
# 2012-02-10 (CW) Added journal name and PMID (v2.0).
# 2012-02-14 (CW) Added versioned file names (v2.1).
# 2012-02-15 (CW) Branched development to PyGtk. 
# 2012-02-24 (CW) Major rewrite. End of almost all command line params. Config
#   file instead. Preferences dialog based on config file started.
# 2012-02-29 (CW) ED edits. Coot window prefs in config, Save as... dialog and
#   user feedback on currently edited file name in window title (v0.91).
# 2012-03-02 (CW) Table load by menu bar action. Removed legacy parameters and
#   refactored code a bit. (v.0.92)
# 2012-03-07 (CW) All config file parameters now fully configurable from the
#   preferences menu. Coot temporary directory became a config file parameter.
#   (v0.93).
# 2012-03-08 (CW) Tooltips for Preferences window. (v0.94)
# 2012-03-14 (CW) Added Python/Scheme radiobutton and Ed's coot script. (v0.95)
# 2012-03-15 (CW) Sorted list of het groups all over, facilitates orientation.
#   (v0.96)
# 2012-03-23 (CW) Partial undo of latest sort function: Keep order as in table
#   for jumping forth and back in coot. (v0.97)
# 2012-03-27 (CW) Removed parameter class. (v0.98)
# 2012-05-02 (CW) Added GPLv3 license and About dialog (v0.99)
# 2012-05-02 (CW) Added main table mouse overs, first version that appears to
#   be done. (v1.00)
# 2012-05-09 (CW) Removed bug resulting from a sorting order of a non-Rank
#   column. Altered Assign menu entry names. (v1.01)
# 2012-05-14 (CW) Added release date and refinement software columns. Made
#   columns resizable. Software column fixed width. Border width unified.
#   (v1.02)
# 2012-05-21 (CW) Smarter adding of user stamp. If the comment already starts
#   with the user's abbreviated name, it is not prepended. (v1.03)
# 2012-06-06 (CW) Some flags for figure preparation mode (v1.04).
# 2012-06-14 (CW) Fixed journal column to show full journal name. (v1.05).
# 2012-06-24 (CW) Baptized it Twilight (v1.06).
# 2012-07-27 (CW) Manual loading of arbitrary PDB entries with ligands. (v1.07)
# 2012-07-30 (CW) Bug removed when importing PDB entries on Rank sorted column.
# 2012-07-31 (CW) <return> key loads structure in import widget.
# 2012-08-01 (CW) PdbFile instance always kept in sync with the config. (v1.08)
# 2012-08-02 (CW) Bug for Rank sorted columns and import remained. Fix. (v1.09)
# 2012-10-07 (CW) Empty table at startup and New in File menu. Error handling
#   when writing to a non-writable file. 
# 2012-10-10 (CW) More informative error messages when retrieving entries.
# 2012-10-14 (CW) Rewrote meaning of bad ligands. Good and bad ones, both can
#   now be shown in the viewer, though only exclusively. Group sort by
#   Combined Score or RSCC. Good and bad RSCC coloring of ligands. Support for
#   color vision deficiencies. (v1.10)  
# 2012-10-23 (CW) Highlight the selected ligand in the ligand table,
#   irrespective of good/bad viewing mode. (v1.11)
#
# known bugs:
#
#


###########
###########
### TOO DOO
###########
###########

# Switch to showing only valid ligands in Ed's scheme coot script.

import os
import sys
import gzip
import argparse
import subprocess
import webbrowser
try:
    import pygtk
    pygtk.require('2.0')
    import gtk
    import gobject;
except ImportError:
    sys.stderr.write("Could not import PyGtk. Please check installation.\n");
    sys.exit(1);

from misc import CalcCombinedScore, EMPTYFN, UNTITLEDFN, NONLIGANDS,\
     RSCCBADTH, RSCCGOODTH;
from edstable import CEDSTableData;
from pdbfile import CPDBFile, PDB_OK, PDB_NO_CONNECTION,\
     PDB_PDBENTRY_NOT_EXISTS, PDB_EDSENTRY_NOT_EXISTS,\
     PDB_EDS_INVALID_STATLIS,PDB_NO_LIGANDS_FOUND;

from files import GetVersionedInOutFileNames;
from optconf import CDysConfigurer;

# Program version
__version__         = "1.11";

# Main window title
MAIN_WINDOW_TITLE = "Ligand Dysplayer";
# Preferences window title.
PREF_WINDOW_TITLE = "Preferences";
# Import EDS and PDB window.
IMPORTEDSPDB_WINDOW_TITLE = "Import EDS and PDB";
# Load dialog
LOAD_TITLE = "Load table...";
# Save as... dialog
SAVE_AS_TITLE = "Save table as...";
# File choosing dialog for coot window title.
COOT_FILE_SEL_WINDOW_TITLE = "Choose coot executable";
# File choosing dialog for coot window title.
PDB_DIR_SEL_WINDOW_TITLE = "Select PDB mirror directory";
# Same for PyMOL
PYMOL_FILE_SEL_WINDOW_TITLE = "Choose PyMOL executable";

# File is not saved dialog.
TABLE_MODIFIED_CONTINUE_TEXT = "Table has been modified. Continue anyway?";
# Invalid input file format
INVALID_FILE_FORMAT_TEXT = "Invalid file format. No data has been read.";
# No electron density data found on EDS server.
NO_EDS_DATA_AVAILABLE_TEXT = "No EDS entry available for PDB code ``%s''.";
# There is no ligand in the PDB file.
NO_LIGAND_DATA_AVAILABLE_TEXT = "PDB code ``%s'' contains no ligands.";
# PDB entry is not available on PDB server, ie. invalid PDB code.
INVALID_PDBENTRY_TEXT = "PDB entry ``%s'' does not exist on RCSB PDB server (Invalid PDB code)."
# (coot) file is not executable.
FILE_NOT_EXECTUTABLE_TEXT="File is not executable."
# Somehow this file is not there.
FILE_NOT_EXIST_TEXT="File does not exist."
# Not a directory has been selected.
NOT_A_DIRECTORY_TEXT="Not a directory."
# Directory cannot be cd'd.
CANNOT_CHG2DIRECTORY_TEXT="Cannot change to directory."
# Can't write to dir.
NO_WRITE_PERMS_TEXT= "No write permission on directory."
# Can't write to file.
COULD_NOT_WRITE_FILE_TEXT = "Could not write file ``%s''. Permissions ok?";
# URL could not be opened.
CONNECTION_FAILED_TEXT = "Connection problem. Could not fetch data from server.";

# File menu entries
FILE_MNU_FETCH_EDS_TEXT = "Fetch EDS and PDB...";

# The license text as it appears in the About window.
LICENSE_TEXT = """Twilight %s

Twilight is a GUI for displaying ligands incompatible with electron density.

Copyright (C) 2012  Christian X. Weichenberger, Edwin Pozharski, and Bernhard Rupp.

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program.  If not, see <http://www.gnu.org/licenses/>.
""" % __version__;

# Tooltip texts
# These are for the Preferences window.
CONTOUR_2FOFC_TTEXT = "Contour level in units of standard deviations the \
2FoFc map is drawn on coot startup.";
CONTOUR_FOFC_TTEXT = "Contour level in units of standard deviations the \
FoFc map is drawn on coot startup."
COOT_WIN_WIDTH_TTEXT = "Coot window width after launching."
COOT_WIN_HEIGHT_TTEXT = "Coot window height after launching."
COOT_BINARY_TTEXT = "Full path file name of coot executable, in case it is \
not in the default search path. If it is, simply put in the command line \
executable name in here. This is either 'coot' on Unix-based systems or \
'runwincoot.bat' on MS-Windows.";
PDB_MIRROR_DIR_TTEXT = "Directory name of a local PDB mirror for retrieval \
of coordinate files without connecting to the PDB server. This directory \
name is the root of a tree that usually is generated by rsyncing with the \
PDB server."
TEMP_FILES_DIR_TTEXT = "Directory for storing downloaded coordinate and \
electron density files. This directory needs to have write permissions. \
Files are kept in between sessions. Structure files are loaded from here \
before connecting to the PDB or the EDS server for file retrieval.";
USER_ID_TTEXT = "Short name of the annotator. This will automatically \
be prefixed to the annotation text."
SCRIPT_PYTHON_TTEXT = "Use coot's Python scripting engine for exchanging \
data (recommended).";
SCRIPT_SCHEME_TTEXT = "Use coot's Scheme scripting engine for exchanging \
data (currently does not implement full functionality.";
CONFIG_FILENAME_TTEXT = "File name where all configuration options entered \
in this window are saved to. You can manually edit this file, especially \
if you want to rephrase or extend the annotation texts.";
# These are for the Import from EDS dialog.
EDS_PDB_ENTRY_TTEXT = "PDB four letter code for retrieval of a structure \
and its corresponding electron density from EDS.";

RANK_TTEXT = "Rank shows highest Score values first and lists below all other \
ligands from the same PDB entry that have passed the filter criteria.";
PDBID_TTEXT = "PDB four letter code.";
LIGNM_TTEXT = "Ligand three letter residue name.";
RESNR_TTEXT = "Chain name, residue number and insertion code.";
RSCC_TTEXT = "Real space correlation coefficient.";
OWAB_TTEXT = "Occupancy-weighted average B-factor.";
MOLWT_TTEXT = "Molecular weight of ligand.";
CHARGE_TTEXT = "Overall charge of ligand.";
NRHEAVY_TTEXT = "Number of heavy atoms of ligand.";
NRHYDRO_TTEXT = "Number of hydrogen atoms of ligand.";
NRATOMS_TTEXT = "Overall number of ligand atoms, heavy atoms and hydrogens.";
RELDAT_TTEXT = "Deposit date of PDB entry.";
RESOL_TTEXT = "Experimental resolution of the structure."
RWORK_TTEXT = "R-value of the working set.";
RFREE_TTEXT = "R-value of the testing set.";
ROBS_TTEXT = "R-value of both the test and working sets.";
SW_TTEXT = "Software used during refinement of crystal structure.";
JRNL_TTEXT = "Abbreviation of the journal the structure was published in.";
PMID_TTEXT = "Numeric PubMed identfier for the publication, if available. \
Otherwise, a hyphen (-) indicates that there is no publication.";
CLASS_TTEXT = "Classification code of the ligand. 'B' signifies it is a \
frequently used buffer molecule.";
VALID_TTEXT = "Currently this column is always set to yes (Y).";
SCORE_TTEXT = "Score combining Resol and RSCC into a single number.";
COMMENT_TTEXT = "Author's annotation.";

MAIN_TABLE_TOOLTIP_TEXTS = [PDBID_TTEXT, LIGNM_TTEXT, RESNR_TTEXT,
                            RSCC_TTEXT, OWAB_TTEXT, MOLWT_TTEXT, CHARGE_TTEXT,
                            NRHEAVY_TTEXT, NRHYDRO_TTEXT, NRATOMS_TTEXT,
                            RELDAT_TTEXT, RESOL_TTEXT, RWORK_TTEXT,
                            RFREE_TTEXT, ROBS_TTEXT, SW_TTEXT, JRNL_TTEXT,
                            PMID_TTEXT, CLASS_TTEXT, VALID_TTEXT, SCORE_TTEXT,
                            COMMENT_TTEXT]; 
# Switch for manuscript figure display settings
PREPARE_FIGURE_MODE = False;

# Number of columns for the ligand array
LIG_TABLE_MAX_COL = 10;
# Font size (point) of the monospace fixed font for the two tables.
TABLES_FONT_SIZE_PTS = 8;

# Border width of the main views EDS Table, Ligand detail viewer and Status line
BORDER_WIDTH   = 7;
# Minimal width of any resizable column
MIN_COL_WIDTH  = TABLES_FONT_SIZE_PTS*3;
# Width of the Softwre column, too large by default.
SW_COL_WIDTH   = TABLES_FONT_SIZE_PTS*11;
# Width of the Jrnl column, too large by default.
JRNL_COL_WIDTH = TABLES_FONT_SIZE_PTS*10;


# Colors for hetgroup display table.
LOW_RSCC_COL     = "#f1641d";
IN_FOCUS_BG_COL  = "#cccccc";
HIGHLIGHT_FG_COL = "#ffffff";
HIGHLIGHT_BG_COL = "#4b6983";
BLACK_COL        = "#000000";
HIGH_RSCC_COL    = "#00a277";
DEFAULT_COL      = None;

# Colors for distinguishing manually added structures from those in the
# precompiled table.
VALID_ROW_COL          = None;
MANUALLY_ADDED_ROW_COL = "#d0e1ff";

# Status messages.
STAT_FETCHING_PDB_ENTRY = "Fetching PDB entry from RCSB PDB server.";
STAT_FETCHING_EDS_DATA  = "Fetching electron density data from EDS server.";
STAT_LAUNCHING_COOT     = "Launching coot...";
STAT_SAVING_FILE        = "Saving file ``%s''...";
STAT_LOADING_FILE       = "Loading file ``%s''..."

# Timeout for clearing status line (ms)
CLR_TIMEOUT = 1500;

# Callback globals
MNU_SEND_TO_PDB     = "sendPDB";
MNU_SEND_TO_EDS     = "sendEDS";
MNU_SEND_TO_PMED    = "sendPubMed";
MNU_SEND_TO_PDBREDO = "sendPDB_REDO";

# URLs for looking up data from a table row.
URL_PDB     = "http://www.pdb.org/pdb/explore/explore.do?structureId=%s"
URL_EDS     = "http://eds.bmc.uu.se/cgi-bin/eds/uusfs?pdbCode=%s"
URL_PDBREDO = "http://www.cmbi.ru.nl/pdb_redo/%s/%s/index.html"
URL_PUBMED  = "http://www.ncbi.nlm.nih.gov/pubmed/%s"

class CGtkDysplayer:
    """GUI for analyzing ligands not compatible with their respective electron
    density.
    """
    def __init__(s, tableFN, configFN, sortS=True, badLigands=True):
        """Constructor.
        
        @param tableFN  EDS table file name, EMPTYFN indicates empty table.
        @param configFN  Configuration file name.
        @param sortS  @c True, if we complex sort by score S, @c if we sort by
         RSCC only.
        @param badLigands: @c True, show the list of edns-incompatible
         ligands, else show the highly compatible ones.
        """
        s._EDSInTableFN = tableFN;
        if( configFN!="" ):
            # have a config file name to load or create other than the default.
            dirNm, confFN = os.path.split(configFN);
            s._DysConfig = CDysConfigurer(defaultarget=dirNm,
                                          cfgname=confFN);
        else:
            # default config name and location.
            s._DysConfig = CDysConfigurer();
        cfg = s._DysConfig.GetConfig();
        s._Table = CEDSTableData();
        if( sortS ):
            s._Table.SetSortingScheme(s._Table.SCOREIDX, False);
        else:
            s._Table.SetSortingScheme(s._Table.RSCCIDX, True);
        if( badLigands ):
            s._Table.SetFocusOnBad();
        else:
            s._Table.SetFocusOnGood();
        s._PdbFile = CPDBFile();
        s._UpdateFromConfig();
        s.populatedPDBID = "";
        s.tableModified = False;
        s.__BuildWindow();
        s._UpdateListViewWithTableFile(s._EDSInTableFN)
        s.preferencesWin = CDysPreferences(s._DysConfig);
    
    def _GetTableFileName(s):
        """Get the 'translated' file name for the table. This takes care of
        the empty table and its associated Untitled file name.
        """
        if( s._EDSInTableFN==EMPTYFN ):
            return UNTITLEDFN;
        else:
            return s._EDSInTableFN;
    
    def _UpdateFromConfig(s):
        """Update the internal PdbFile instance with the current setting from
        the config.
        """
        cfg   = s._DysConfig.GetConfig();
        s._PdbFile.SetPDBDir(cfg.get("program", "pdbDir"));
        s._PdbFile.SetCootTmpDir(cfg.get("program", "tempDir"));
        s._PdbFile.SetCootExec(cfg.get("program", "cootExec"));
        
    def __BuildWindow(s):
        """Constructs main window.
        """
        # Miscellaneous stuff for setting up the framework of the main
        # application window.
        s.window = gtk.Window(gtk.WINDOW_TOPLEVEL);
        s.window.connect("destroy", s._QuitCB);
        s.window.connect("delete_event", s._QuitCB);
        s.window.set_title(s.GetMainWindowTitle());
        s.window.set_border_width(0);
        s.window.set_size_request(1000, 750);
        
        # We pack everything into a vertical box.
        s.mainBox = gtk.VBox(False, 0);
        s.window.add(s.mainBox);

        # First the menu bar.
        s._BuildMenuBar();
        s._BuildPopupMenu();
        s.mainBox.pack_start(s.menuBar, False, False, 0);
        
        # Create the data storage for the ligand table.
        # Number of columns in the stored EDS table object s._Table.
        nrTableColumns = len(s._Table.ColumnTypes);
        # the large table with all ligand data.
        s.edsTableWin = gtk.ScrolledWindow();
        s.edsTableWin.set_border_width(BORDER_WIDTH);
        s.mainBox.pack_start(s.edsTableWin, True, True, 0);
        # Column types:
        # * display data
        # * data in natural data types for sorting
        # * control data
        columnTypes = len(s._Table.ColumnTypes)*[str ,] + \
                      list(s._Table.ColumnTypes) + [int, str]
        # these are the two locations where the comments go.
        s.commentIdcs = (s._Table.COMMENTIDX,
                         nrTableColumns+s._Table.COMMENTIDX);
        s.edsLS = gtk.ListStore(*columnTypes);
        s.edsLS.clear();
        
        # Create the actual gtk GUI element that renders the table. Connect it
        # with the ListStore object.
        s.edsView = gtk.TreeView(s.edsLS);
        s.edsView.connect("button-press-event", s._TableButtonPressCB);
        # Two types of columns. One general and another that allows input for
        # the editing process.
        s.rendTxt = gtk.CellRendererText();
        s.rendTxt.set_property("family", "monospace");
        s.rendTxt.set_property("ypad", 0);
        s.rendTxt.set_property("size-points", TABLES_FONT_SIZE_PTS);
        # for the figure in the manuscript:
        if( PREPARE_FIGURE_MODE ):
            s.rendTxt.set_property("font", "monospace bold 13");
        s.editTxt = gtk.CellRendererText();
        s.editTxt.set_property("family", "monospace");
        s.editTxt.set_property("ypad", 0);
        s.editTxt.set_property("size-points", TABLES_FONT_SIZE_PTS);
        s.editTxt.set_property("editable", True);
        if( PREPARE_FIGURE_MODE ):
            s.editTxt.set_property("font", "monospace bold 13");
        s.editTxt.connect('edited', s._CommentCB)
        
        # add_with_viewport would scroll the table header.
        s.edsTableWin.add(s.edsView);
        
        # All ligands per PDB entry TreeView, a table with the color coded
        # ligand residue numbers.
        s.ligTableWin = gtk.ScrolledWindow();
        # Allow arbitrary width, but limit the height to 120px
        s.ligTableWin.set_size_request(-1, 120);
        s.ligTableWin.set_border_width(BORDER_WIDTH);
        # Do not take extra space when enlarging window.
        s.mainBox.pack_start(s.ligTableWin, False, False, 0);
        # these are the raw strings for the hetgroup names in the ListStore.
        columnTypes  =  LIG_TABLE_MAX_COL*[str ,];
        # One time for the foreground color, one time for the background.
        columnTypes += 2*LIG_TABLE_MAX_COL*[str, ];
        s.ligLS = gtk.ListStore(*columnTypes);
        s.ligLS.clear();
        s.ligView = gtk.TreeView(s.ligLS);
        # Get the selection object and disable selection on this table. We do
        # not want to have highlighted rows in the ligand display.
        s.ligView.get_selection().set_mode(gtk.SELECTION_NONE);
        s.ligView.set_grid_lines(gtk.TREE_VIEW_GRID_LINES_BOTH);
        # This time, we do not want to display table headers.
        s.ligView.set_headers_visible(False);
        for tabLoc in range(LIG_TABLE_MAX_COL):
            # Columns are nameless, as we disabled column headers
            col = gtk.TreeViewColumn("", s.rendTxt, text=tabLoc,
                                     foreground=LIG_TABLE_MAX_COL+tabLoc,
                                     background=2*LIG_TABLE_MAX_COL+tabLoc);
            s.ligView.append_column(col);
        s.ligTableWin.add(s.ligView);

        # Status bar
        s.statusBox = gtk.HBox(False, 0);
        s.statusBox.set_border_width(BORDER_WIDTH);
        s.statusLabel = gtk.Label("Status: ");
        s.statusEntry = gtk.Entry();
        s.statusEntry.set_text("");
        s.statusEntry.set_editable(False);
        s.statusBox.pack_start(s.statusLabel, False, False);
        s.statusBox.pack_start(s.statusEntry, True, True);
        s.mainBox.pack_start(s.statusBox, False, True, 0);
        
        s.window.show_all(); # instead of single widget .show() methods

    def _GetListViewRowColor(s, idx):
        """Determine row color of an entry in the main table.
        @param idx  Index into the EDS table, s._Table.
        """
        rowColor = VALID_ROW_COL;
        row      = s._Table[idx];
        if( row[s._Table.VALIDIDX]=="A" ):
            rowColor = MANUALLY_ADDED_ROW_COL;
        return rowColor;
        
    def _UpdateListViewWithTableFile(s, filename):
        """Update the ListStore and TreeView with new data from another EDS
        table file.
        @param filename  File name of table to read. If set to the emtpy
         string, the default template for an empty table will be read.
        """
        if( filename=="" ):
            s._Table.ReadTable(EMPTYFN);
        else:
            s._Table.ReadTable(filename);
            
        nrTableColumns = len(s._Table.ColumnTypes);
        s.edsLS.clear();
        # The List Store has the following columns:
        # All columns from the EDS table object as strings             THEN
        # the same columns but converted to natural types for sorting  THEN
        # columns used for controlling the table attributes:
        #  o  Original sorting order by a rank integer column.
        for i, row in enumerate(s._Table):
            rowColor = s._GetListViewRowColor(i);
            #       string   natural types         controls
            lsRow = row + s._Table.ConvertRow(i) + [i+1, rowColor];
            s.edsLS.append(lsRow);

        # Empty the view first.
        for column in s.edsView.get_columns():
            s.edsView.remove_column(column);
        # we added hidden control columns at the end of the data store, let's
        # remember its index in the list store.
        s.rankColIdx = 2*len(s._Table.ColumnTypes);
        s.rowColorIdx= s.rankColIdx + 1;
        # We now pack the first visible column, the rank. This column is not
        # included in the EDS table file and therefore is generated here in
        # order to have a way back for the original ranking, which is too
        # complex to redo on the fly.
        s.rankCol = gtk.TreeViewColumn("", s.rendTxt, text=s.rankColIdx,
                                       background=s.rowColorIdx);
        rankLabel = gtk.Label("Rank");
        rankLabel.show();
        rankLabel.set_tooltip_text(RANK_TTEXT);
        s.rankCol.set_widget(rankLabel);
        s.rankCol.set_sort_column_id(s.rankColIdx);
        s.edsView.append_column(s.rankCol);
        # now pack all columns as they appear in the EDS input table, using
        # strings to display since float data types are displayed with too many
        # digits and waste space.
        toolTips = gtk.Tooltips();
        for tabLoc, colName in enumerate(s._Table.GetHeader()):
            cellRenderer = s.rendTxt if tabLoc!=s._Table.COMMENTIDX else \
                           s.editTxt;
            col = gtk.TreeViewColumn("", cellRenderer, text=tabLoc,
                                     background=s.rowColorIdx);
            # sorting is done through an offset column with the same data but
            # converted to its natural data type.
            if( not PREPARE_FIGURE_MODE ):
                # we want sortable columns. But in figure preparation the sort
                # order arrow space is wasted and renders the table too
                # large.
                col.set_sort_column_id(tabLoc + nrTableColumns);
            col.set_resizable(True);
            col.set_min_width(MIN_COL_WIDTH);
            # The software column is too wide. Narrow manually here.
            if(   tabLoc==s._Table.SWIDX ):
                col.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED);
                col.set_fixed_width(SW_COL_WIDTH);
            elif( tabLoc==s._Table.JRNLIDX ):
                col.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED);
                col.set_fixed_width(JRNL_COL_WIDTH);
                
            colLabel = gtk.Label(colName);
            colLabel.show();
            colLabel.set_tooltip_text(MAIN_TABLE_TOOLTIP_TEXTS[tabLoc]);
            col.set_widget(colLabel);
            s.edsView.append_column(col);

        # Data has been loaded, therefore, it is ok to save now, too.
        s._EnableSaveMenus()
        
        
    def _ShowPDBEDSErrorMessage(s, errorCode, pdbID=""):
        """Given an error message from the @c pdbfile module, pop up an
        informative text box.
        @param errorCode  Error code as defined in pdbfile.py
        @param pdbID  PDB entry if available. Will be printed for certain msgs.
        """
        if(   errorCode==PDB_NO_CONNECTION ):
            # Connection to the internet could not be established.
            SimpleOKDialog(CONNECTION_FAILED_TEXT);
        elif( errorCode==PDB_PDBENTRY_NOT_EXISTS ):
            # Could not fetch PDB file (this is an invalid PDB code then)
            SimpleOKDialog(INVALID_PDBENTRY_TEXT % pdbID);
        elif( errorCode==PDB_EDSENTRY_NOT_EXISTS ):
            # The PDB file was fetched, but the MAPMAN statistics list file
            # has not been retrieved from the EDS server. This means that
            # EDS has not generated an electron density for this PDB entry.
            SimpleOKDialog(NO_EDS_DATA_AVAILABLE_TEXT % pdbID);
        elif( errorCode==PDB_NO_LIGANDS_FOUND ):
            # Everything fine, but there was no ligand in the PDB file.
            SimpleOKDialog(NO_LIGAND_DATA_AVAILABLE_TEXT % pdbID);
        elif( errorCode==PDB_EDS_INVALID_STATLIS ):
            sys.stderr.write("Tried to fetch EDS data for another PDB file (internal error)");
    
    
    def _PopulateLigandTable(s, tableIdx):
        """Fill the second table with ligand data and color it.
        @param tableIdx  Index into the Gtk liststore object. Not to be
         confused with the EDS table object.
        @return @c True, ok. @c False, something went wrong. Error message has
         been displayed. 
        """
        # The rank column is our save way back to the s._Table indices.
        rowIdx    = s.edsLS[tableIdx][s.rankColIdx] - 1;
        # this could also be retrieved directly from the eds ListStore...
        pdbID     = s._Table.GetPDBID(rowIdx);
        ligRnr    = s._Table.GetResNr(rowIdx);
        # but for this here, we need to row index in s._Table anyway!
        validRnrs = s._Table.GetAllDisplayLigands(rowIdx);
        focRnrs   = s._Table.GetAllRSCCFocusLigands(rowIdx);
        s._Table.ToggleFocus();
        altFocRnrs= s._Table.GetAllRSCCFocusLigands(rowIdx);
        s._Table.ToggleFocus();
        # Get the latest content of the config and apply it to PdbFile.
        s._UpdateFromConfig();
        s.SetStatusText(STAT_FETCHING_PDB_ENTRY);
        status = s._PdbFile.FetchPDB(pdbID);
        s.DelayedStatusClear();
        if( status!=PDB_OK ):
            s._ShowPDBEDSErrorMessage(status, pdbID=pdbID);
            return False;
        s.populatedPDBID = pdbID;
        s.ligLS.clear();
        if( s._Table.IsFocusOnBad() ):
            focCol, altFocCol = LOW_RSCC_COL, HIGH_RSCC_COL;
        else:
            focCol, altFocCol = HIGH_RSCC_COL, LOW_RSCC_COL;
        row, fgCols, bgCols = [], [], [];
        for rnr, ligand in s._PdbFile.GetAllHetGroups():
            resNm = ligand.Get3LetterName();
            # do not show HOH et al.
            if( resNm in NONLIGANDS ):
                continue ;
            fgCol = bgCol = DEFAULT_COL;
            # Coloring has become quite complex. First, if the entry has been
            # imported manually, the rules are different. Color the good and
            # the bads, do not use bg colors, as everything is displayed
            # anyway. 
            # If the entry is present in the table and has not been imported,
            # it makes sense to color by presence in the table, which has the
            # meaning that the present ligands are either good or bad.
            # In either case, highlight the selected ligand.
            if( s._Table.IsManuallyImported(rowIdx) ):
                if( rnr in focRnrs ):
                    fgCol, bgCol = focCol, DEFAULT_COL;
                if( rnr in altFocRnrs ):
                    fgCol, bgCol = altFocCol, DEFAULT_COL;
            else:
                # A ligands can be both in focus AND be valid (ie. G/Y).
                # Therefore, assign first the color for focus and later it may
                # be overridden by a valid ligand.
                if( rnr in focRnrs ):
                    fgCol, bgCol = focCol, DEFAULT_COL;
                if( rnr in validRnrs ):
                    fgCol, bgCol = focCol, IN_FOCUS_BG_COL;
                if( rnr in altFocRnrs ):
                    fgCol, bgCol = altFocCol, DEFAULT_COL;
            if( rnr==ligRnr ):
                fgCol, bgCol = HIGHLIGHT_FG_COL, HIGHLIGHT_BG_COL;
            row.append("%s %s" % (resNm, rnr));
            fgCols.append(fgCol);
            bgCols.append(bgCol);
            if( len(row)==LIG_TABLE_MAX_COL ):
                s.ligLS.append(row+fgCols+bgCols);
                row, fgCols, bgCols = [], [], [];
        if( row!=[] ):
            filler = (LIG_TABLE_MAX_COL-len(row))*[DEFAULT_COL,];
            row    = row + filler;
            fgCols = fgCols + filler;
            bgCols = bgCols + filler;
            s.ligLS.append(row+fgCols+bgCols);
        return True;
        
    def _CommentCB(s, cell, path, newText):
        """Callback function for editing the Comment column.
        'cell' argument is ignored!"""
        # we update both the original column used for display and the mirrored
        # column with the natural data type, which happens to be the same
        # (str) for the Comment column.
        # EP: newText is prepended with user stamp
        for colIdx in s.commentIdcs:
            ustamp = s._DysConfig.GetConfig().get("program", "ustamp").strip();
            if( ustamp!="" and newText.strip().find(ustamp)!=0 ):
                newText = ustamp + " " + newText;
            s.edsLS[path][colIdx] = newText;
        s.SetModifiedContent();

    def _SaveFile(s, outFFN):
        """Saves the currently displayed data into a file.
        The file name has been determined at startup. This method retrieves
        the comments, transfers them back to the EDSTable object and writes
        out the modified content.
        
        @param outFFN  Full path output file name.
        @return @c True, everything fine. @c False, some error during
         writing. An info dialog has already been displayed.
        """
        for row in s.edsLS:
            # the index in the originally read table is given by the rank
            # column, which is displayed as the first column, but was appended
            # to the table in the ListStore.
            origIdx = row[s.rankColIdx] - 1;
            # this text comes from the edits.
            comment = row[s._Table.COMMENTIDX];
            s._Table[origIdx][s._Table.COMMENTIDX] = comment;
        ok = s._Table.WriteTSV(outFFN, verbose=True);
        if( not ok ):
            OkInfoDialog(COULD_NOT_WRITE_FILE_TEXT % outFFN);
        return ok;

    def _NewCB(s, widget, data=None):
        """Callback for creating an empty table."""
        s._LoadCB(widget, data=data, empty=True);
    
    def _LoadCB(s, widget, data=None, empty=False):
        "Callback for Load table dialog."
        if( s.tableModified ):
            yesPressed = YesNoInfoDialog(TABLE_MODIFIED_CONTINUE_TEXT);
            if( not yesPressed ):
                # ok, user is afraid, let's not load anything this time.
                return ;
        
        if( not empty ):
            inFFN = HandleFileChooser(LOAD_TITLE, gtk.FILE_CHOOSER_ACTION_OPEN);
        else:
            # Empty table template file name.
            inFFN = EMPTYFN;
        if( inFFN!=None ):
            # First, let's check if this stuff we are being fed is a valid EDS
            # table...
            if s._Table.IsValidTableFile(inFFN):
                s.SetStatusText(STAT_LOADING_FILE % inFFN);
                s._UpdateListViewWithTableFile(inFFN);
                s._EDSInTableFN = inFFN;
                s.DelayedStatusClear();
                s.ClearModifiedContent();
                s._EnableFetchEDSnPDB();
            else:
                # no it isn't valid, complain and leave it.
                OkInfoDialog(INVALID_FILE_FORMAT_TEXT);
        
    def _LoadSfCB(s, widget, data=None):
        """Callback for the Fetch EDS and PDB dialog."""
        importDlg = CImportEDSnPDB();
        retVal    = importDlg.run();
        pdbID     = importDlg.GetPDBCode();
        importDlg.destroy();
        if( retVal!=gtk.RESPONSE_OK ):
            # Cancel or window closed. Do not proceed.
            return ;
        # Get the latest content of the config and apply it to PdbFile.
        s._UpdateFromConfig();
        s.SetBusyCursor()
        s.SetStatusText(STAT_FETCHING_PDB_ENTRY);
        status = s._PdbFile.FetchPDBnStatLis(pdbID);
        s.DelayedStatusClear();
        s.SetNormalCursor();
        if( status!=PDB_OK ):
            s._ShowPDBEDSErrorMessage(status, pdbID=pdbID);
            return ;
        
        # these are the ligands we need to add.
        ligandsToAdd = s._PdbFile.GetAllHetGroups();
        nrToAdd      = len(ligandsToAdd);
        # Compute the combined score and sort by it. This reflects the
        # behavior as expected from the group sorting.
        for rnr, ligand in ligandsToAdd:
            score = CalcCombinedScore(ligand.GetRSCC(),
                                      s._PdbFile.GetResolution());
            ligand.SetCombinedScore(score);
        # Sort descending by combined score.
        ligandsToAdd.sort(lambda x,y: -(x[1].GetCombinedScore()<y[1].GetCombinedScore()));
        # Since the list store reflects the sort order of the tree view, we
        # need to compute the new ranks. Fortunately, we are adding new
        # entries to the begin of the list, so we know that the current ranks
        # need to be incremented by the number of entries added.
        #
        # Word of warning: Do not enumerate(s.edsLS) and manipulate rows at
        # the same time. For example, this code goes into an infinite loop
        # when enumerating and altering the rank column if the table is sorted
        # by this column!
        #
        # It's getting worse than that: If sorting by Rank is on, and sorting
        # is ascending, it will not work with directly indexing into the list
        # store, either. It seems the list store is immediately resorting its 
        # content once an element has been modified! If sorting is on the Rank
        # column, modification of the Rank cells has to be executed such that
        # the sorting will not be overthrown.
        # Example
        # rank = [10,11,12,13]
        # Adding the integer to the top entry in the list (10) will yield
        # rank = [11,12,12,13] after '10' has been added by '2', since the list
        # store immediately restores sort order.
        # Adding from the bottom will not destroy the order:
        # (add 2 to position 3 where integer 13 is): [10,11,12,15]
        # (add 2 to position 2 where integer 12 is): [10,11,14,15]
        # (add 2 to position 1 where integer 11 is): [10,13,14,15]
        # (add 2 to position 0 where integer 10 is): [12,13,14,15]
        #
        # It's easy to see that the same argument holds if the list is sorted
        # descending when accessing the elements from top to bottom.
        #
        # Solution: Always increment from the point of the list where there is
        # the largest element. If this is increased, it will not change
        # position. (If the list store is not sorted by Rank, no position
        # change will happen at all, since we are not operating on the sort
        # column.)
        #
        # We do not care if the list store is currently sorted by the Rank
        # column, we just have a look in which direction the sort order could
        # be and then choose the right direction for accessing the elements of
        # the list store.
        if( len(s.edsLS)>0 ):
            order = range(len(s.edsLS));
            if( s.edsLS[0][s.rankColIdx]<s.edsLS[1][s.rankColIdx] ):
                # If sorted by Rank, it is an ascending order. The highest
                # Rank is at the bottom of the table, that's where we want to
                # start then. We traverse the elements for updating from the
                # bottom to the top.
                order.reverse();
            for i in order:
                s.edsLS[i][s.rankColIdx] = s.edsLS[i][s.rankColIdx] + nrToAdd;
        # Add each new ligand as a single new row and also set the rank.
        index = nrToAdd;
        for rnr, ligand in ligandsToAdd:
            # Add it to the EDS table object.
            s._Table.AddLigand(pdbID, s._PdbFile, ligand);
            # Determine the color of the freshly added row.
            rowColor = rowColor = s._GetListViewRowColor(0);
            # And once there, simply take it and add it to the list store.
            lsRow  = s._Table[0] + s._Table.ConvertRow(0) + [index, rowColor];
            s.edsLS.prepend(lsRow);
            index -= 1;
        # New data has been added, time to think about saving that stuff.
        s.SetModifiedContent();
        # This is a bit inconsistent: Jump to the top of the table,
        # independent of the sorting order. If the sorting order is by
        # combined score/rank, this will result in the newly loaded
        # structure.
        s.edsView.scroll_to_cell(0);
        
        
    def _PrefCB(s, widget):
        "Callback for Preferences dialog."
        s.preferencesWin.show_all();
    
    def _SaveCB(s, widget, data=None):
        """Callback for saving the table contents."""
        saveFFN = s._GetTableFileName();
        s.SetStatusText(STAT_SAVING_FILE % saveFFN);
        ok = s._SaveFile(saveFFN);
        if( ok ):
            s.DelayedStatusClear();
            s.ClearModifiedContent();

    def _SaveAsCB(s, widget, data=None):
        "Callback function for Save as..."
        curFFN = s._GetTableFileName();
        edsInFN, edsOutFN = GetVersionedInOutFileNames(curFFN);
        outFFN = HandleFileChooser(SAVE_AS_TITLE, gtk.FILE_CHOOSER_ACTION_SAVE,
                                   suggestedFN=edsOutFN);
        if( outFFN!=None ):
            s.SetStatusText(STAT_SAVING_FILE % outFFN);
            s._SaveFile(outFFN);
            # This is a trick: The new input file is the chosen save file. By
            # this, save will default to this new file. Also, the window title
            # adjusts.
            s._EDSInTableFN = outFFN;
            s.DelayedStatusClear();
            s.ClearModifiedContent();

    def _AboutCB(s, widget):
        """Callback function for About..."""
        SimpleOKDialog(LICENSE_TEXT);
        
    def _QuitCB(s, widget, data=None):
        """Callback for application shutdown. Check if table has been
        saved.
        """
        yesPressed = True;
        if( s.tableModified ):
            yesPressed = YesNoInfoDialog(TABLE_MODIFIED_CONTINUE_TEXT);
        if( yesPressed ):
            gtk.main_quit();
        # We need to return True here since:
        # a) if we did not quit the application, the framework needs to know
        #    that we have taken care of the delete event.
        # b) in case we quit the app, this dialog should not show up again.
        return True;
    
    def _TableMenuCB(s, token):
        """Callback for selected menu item in the table."""
        if( token in (MNU_SEND_TO_PDB, MNU_SEND_TO_EDS, MNU_SEND_TO_PMED,
                      MNU_SEND_TO_PDBREDO) ):

            pdbID = s.edsLS[s.__tmpPath][s._Table.PDBIDIDX];
            pmID  = s.edsLS[s.__tmpPath][s._Table.PMIDIDX];
            mid   = pdbID[1:3];
            if(   token==MNU_SEND_TO_PDB ):
                url = URL_PDB % pdbID;
            elif( token==MNU_SEND_TO_EDS ):
                url = URL_EDS % pdbID;
            elif( token==MNU_SEND_TO_PMED ):
                url = URL_PUBMED % pmID;
            elif( token==MNU_SEND_TO_PDBREDO ):
                url = URL_PDBREDO % (mid,pdbID);
            webbrowser.open(url);
        else:
            txt = s._DysConfig.GetConfig().get("annotation", token);
            txt = txt.split("-")[1];
            txt = txt.strip();
            s._CommentCB(None, s.__tmpPath, txt);
        
    def _BuildMenuBar(s):
        """Build and pack the (small) menu bar."""
        aclGrp = gtk.AccelGroup();
        s.window.add_accel_group(aclGrp);
        
        s.menuBar      = gtk.MenuBar();
        s.fileMBMenu   = gtk.Menu();
        s.fileMBItem     = gtk.MenuItem("_File");
        s.menuBar.append(s.fileMBItem);
        s.fileMBItem.set_submenu(s.fileMBMenu);
        s.fileMnNewItem  = gtk.ImageMenuItem(gtk.STOCK_NEW);
        s.fileMnLoadItem = gtk.ImageMenuItem(gtk.STOCK_OPEN);
        s.fileMnLdSfItem = gtk.ImageMenuItem(gtk.STOCK_OPEN);
        s.fileMnSaveItem = gtk.ImageMenuItem(gtk.STOCK_SAVE);
        s.fileMnSvAsItem = gtk.ImageMenuItem(gtk.STOCK_SAVE_AS);
        s.fileMnPrefItem = gtk.ImageMenuItem(gtk.STOCK_PREFERENCES);
        s.fileMnAbotItem = gtk.ImageMenuItem(gtk.STOCK_ABOUT);
        s.fileMnQuitItem = gtk.ImageMenuItem(gtk.STOCK_QUIT);
        s.fileMnNewItem.connect("activate", s._NewCB);
        s.fileMnLoadItem.connect("activate", s._LoadCB);
        s.fileMnLdSfItem.connect("activate", s._LoadSfCB);
        s.fileMnSaveItem.connect("activate", s._SaveCB);
        s.fileMnSvAsItem.connect("activate", s._SaveAsCB);
        s.fileMnPrefItem.connect("activate", s._PrefCB);
        s.fileMnAbotItem.connect("activate", s._AboutCB);
        s.fileMnQuitItem.connect("activate", s._QuitCB);
        s.fileMnLdSfItem.set_label(FILE_MNU_FETCH_EDS_TEXT);
        key, mod = gtk.accelerator_parse("<Control>N");
        s.fileMnNewItem.add_accelerator("activate", aclGrp, key, 
                                         mod, gtk.ACCEL_VISIBLE);
        key, mod = gtk.accelerator_parse("<Control>O");
        s.fileMnLoadItem.add_accelerator("activate", aclGrp, key, 
                                         mod, gtk.ACCEL_VISIBLE);
        key, mod = gtk.accelerator_parse("<Shift><Control>O");
        s.fileMnLdSfItem.add_accelerator("activate", aclGrp, key, 
                                         mod, gtk.ACCEL_VISIBLE);
        key, mod = gtk.accelerator_parse("<Control>S");
        s.fileMnSaveItem.add_accelerator("activate", aclGrp, key, 
                                         mod, gtk.ACCEL_VISIBLE);
        key, mod = gtk.accelerator_parse("<Shift><Control>S");
        s.fileMnSvAsItem.add_accelerator("activate", aclGrp, key, 
                                         mod, gtk.ACCEL_VISIBLE);
        key, mod = gtk.accelerator_parse("<Control>P");
        s.fileMnPrefItem.add_accelerator("activate", aclGrp, key, 
                                         mod, gtk.ACCEL_VISIBLE);

        key, mod = gtk.accelerator_parse("<Control>Q");
        s.fileMnQuitItem.add_accelerator("activate", aclGrp, key, 
                                         mod, gtk.ACCEL_VISIBLE);
        for menuItem in (s.fileMnNewItem, s.fileMnLoadItem,
                         s.fileMnLdSfItem, s.fileMnSaveItem,
                         s.fileMnSvAsItem, s.fileMnPrefItem,
                         gtk.SeparatorMenuItem(), s.fileMnAbotItem,
                         gtk.SeparatorMenuItem(), s.fileMnQuitItem):
            s.fileMBMenu.append(menuItem);
        
    def _DisableFetchEDSnPDB(s):
        """Grey out the Fetch EDS and PDB menu.
        """
        s.fileMnLdSfItem.set_sensitive(False);
        
    def _DisableSaveMenus(s):
        "Grey out Save menus."
        s.fileMnSaveItem.set_sensitive(False);
        s.fileMnSvAsItem.set_sensitive(False);

    def  _EnableSaveMenus(s):
        "Allow Save menus be selected."
        s.fileMnSaveItem.set_sensitive(True);
        s.fileMnSvAsItem.set_sensitive(True);
    
    def _EnableFetchEDSnPDB(s):
        """Allow choosing of Fetch EDS and PDB menu.
        """
        s.fileMnLdSfItem.set_sensitive(True);
        
    def _BuildPopupMenu(s):
        """Build the popup menu structure for the data table."""
        s.tableMenu          = gtk.Menu();
        s.tableMnPDBItem     = gtk.MenuItem("PDB web site");
        s.tableMnEDSItem     = gtk.MenuItem("EDS web site");
        s.tableMnPDBREDOItem = gtk.MenuItem("PDB_REDO web site");
        s.tableMnPMIDItem    = gtk.MenuItem("Pubmed");
        for menuItem, token in ((s.tableMnPMIDItem, MNU_SEND_TO_PMED),
                                (s.tableMnPDBItem, MNU_SEND_TO_PDB),
                                (s.tableMnEDSItem, MNU_SEND_TO_EDS),
                                (s.tableMnPDBREDOItem, MNU_SEND_TO_PDBREDO) ):
            # we need to call show(), otherwise nothing appears when popup()
            # is called.
            menuItem.show();
            menuItem.connect_object("activate", s._TableMenuCB, token);
            s.tableMenu.append(menuItem);
        # sub menu for predefined annotation texts.
        s.tableAsgnSubMenu = gtk.Menu();
        for token, menuDesc in s._DysConfig.GetConfig().items("annotation"):
            menuText = menuDesc.split("-")[0];
            menuText.strip();
            menuItem = gtk.MenuItem(menuText, token);
            menuItem.connect_object("activate", s._TableMenuCB, token);
            menuItem.show();
            s.tableAsgnSubMenu.append(menuItem);
        s.tableAssignItem = gtk.MenuItem("Assign");
        s.tableAssignItem.set_submenu(s.tableAsgnSubMenu);
        s.tableMenu.append(s.tableAssignItem);
        s.tableAssignItem.show()

    def _TableButtonPressCB(s, treeView, event):
        """General callback for all mouse click events happening on the table
        view."""
        # Where did we get this mouse click?
        pthinfo = treeView.get_path_at_pos(int(event.x), int(event.y))
        if( pthinfo is not None ):
            # ok, we really got it somewhere inside the table view.
            path, colObj, cellX, cellY = pthinfo
            pdbID = s.edsLS[path][s._Table.PDBIDIDX];
            cfg   = s._DysConfig.GetConfig();
            # Get the latest content of the config and apply it to PdbFile.
            s._UpdateFromConfig();
            if(   event.button==1 and event.type==gtk.gdk._2BUTTON_PRESS ):
                # this is a left mouse double click, fire up coot.
                resNr = s.edsLS[path][s._Table.RESNRIDX];
                s.SetBusyCursor();
                s.SetStatusText(STAT_FETCHING_PDB_ENTRY);
                # All parameters might have changed since the last double
                # click. Therefore, retrive them from the config and set them
                # in the PDBFile instance.
                status = s._PdbFile.FetchPDB(pdbID);
                if( status!=PDB_OK ):
                    s.SetNormalCursor();
                    s._ShowPDBEDSErrorMessage(status, pdbID=pdbID);
                    return ;
                ok = s._PopulateLigandTable(path[0]);
                if( not ok ):
                    s.SetNormalCursor();
                    return ;
                s.SetStatusText(STAT_FETCHING_EDS_DATA);
                status = s._PdbFile.FetchMTZ(pdbID);
                s.SetNormalCursor();
                s.SetStatusText(STAT_LAUNCHING_COOT);
                s.DelayedStatusClear();
                if( status!=PDB_OK ):
                    s._ShowPDBEDSErrorMessage(status, pdbID=pdbID);
                    return ;
                # this is the trick to get back to the EDS table index when
                # having a different sorting than the default. We use the
                # hidden column which is used by Gtk to define the sorting of
                # the Rank column. Rank numbering starts with 1.
                edsTabIdx = s.edsLS[path][s.rankColIdx] - 1;
                if( s._Table.IsManuallyImported(edsTabIdx) ):
                    displayRnrs = s._Table.GetAllDisplayLigands(edsTabIdx);
                else:
                    displayRnrs = s._Table.GetAllDisplayRSCCFocusLigands(edsTabIdx,
                                                                         False);
                s._PdbFile.SetCootScriptingHost(cfg.get("program",
                                                        "scripting-host"));
                s._PdbFile.WriteCootScript(pdbID, resNr,
                                       cfg.getint("program", "coot-width"),
                                       cfg.getint("program", "coot-height"),
                                       cfg.getfloat("program", "2FoFc"),
                                       cfg.getfloat("program", "FoFc"),
                                       displayRnrs);
                script = s._PdbFile.WriteShellScript(pdbID);
                proc = subprocess.Popen([script,], shell=True, );
            elif( event.button==1 and event.type==gtk.gdk.BUTTON_PRESS ):
                # if we move to a different row, see if the PDB id is still
                # the one we are showing the ligands for. If not, clear the
                # ligand view.
                if( s.populatedPDBID!=pdbID ):
                    s.ligLS.clear();
                    s.populatedPDBID = "";
            elif( event.button==2 and event.type==gtk.gdk.BUTTON_PRESS ):
                # middle mouse button single press, populate the het group
                # data table.
                s.edsView.grab_focus();
                s.edsView.set_cursor(path, colObj, 0);
                # path is [rowNr,]
                s.SetBusyCursor();
                s._PopulateLigandTable(path[0]);
                s.SetNormalCursor();
            elif( event.button==3 and event.type==gtk.gdk.BUTTON_PRESS ):
                # right mouse button single click, build menu.
                activated = s.edsLS[path][s._Table.PMIDIDX]!="";
                s.tableMnPMIDItem.set_sensitive(activated);
                # the table menu callback method will not be furnished with
                # table associated data.
                s.__tmpPath = path;
                # Show the context sensitive menu.
                s.tableMenu.popup(None, None, None, event.button, event.time);

    def GetMainWindowTitle(s):
        "Compose the string for the main window's title."
        ffn = s._GetTableFileName();
        return MAIN_WINDOW_TITLE + ": " + os.path.basename(ffn);
        
    def SetStatusText(s, text):
        """Set a status message in the status bar at the end of the window."""
        s.statusEntry.set_text(text);
        while gtk.events_pending():
            gtk.main_iteration_do(False)

    def DelayedStatusClear(s):
        """Wait a while, then clear the Status line."""
        gobject.timeout_add(CLR_TIMEOUT, lambda self=s:self.SetStatusText(""));
        
    def SetModifiedContent(s):
        """Signify with an asterisk that the table has changed and not yet
        been saved."""
        s.tableModified = True;
        s.window.set_title(s.GetMainWindowTitle() + "*");

    def ClearModifiedContent(s):
        """Signify that the table has not been changed since the last save."""
        s.tableModified = False;
        s.window.set_title(s.GetMainWindowTitle());

    def SetBusyCursor(s):
        """Show an hourglass cursor to indicate work in progress."""
        watch = gtk.gdk.Cursor(gtk.gdk.WATCH);
        s.window.window.set_cursor(watch);

    def SetNormalCursor(s):
        """Set back to cursor to normal."""
        s.window.window.set_cursor(None);


class CDysPreferences(gtk.Window):
    """Handle setting of preferences.

    This window is implemented such that its content can be queried from
    outside. However, as long as the window is mapped, it will report the
    values before its invocation. Only if this window has been closed by
    pressing the OK button, values will be transfered and will be visible for
    the outside caller.
    """
    def __init__(s, dysConfig):
        gtk.Window.__init__(s);
        s.dysConfig = dysConfig;
        s.set_border_width(0);
        s.set_title(PREF_WINDOW_TITLE);
        s.connect("destroy", s._QuitCB);
        s.connect("delete_event", s._QuitCB);
        s.__BuildWindow();

    def __BuildWindow(s):
        # We pack everything into a vertical box.
        s.mainBox = gtk.VBox(homogeneous=False, spacing=5);
        s.add(s.mainBox);
        
        cfg = s.dysConfig.GetConfig();

        # Table row and columns are adjusted to their max element.
        s.Tab1 = gtk.Table(9, 4, homogeneous=False);
        s.sigma2FoFcLabel = gtk.Label("2FoFc contour (sigma): ");
        adj = gtk.Adjustment(cfg.getfloat("program", "2FoFc"),
                             0.00, 50.00, 0.1, -0.1);
        s.sigma2FoFcSpin = gtk.SpinButton(adj, digits=2);
        s.sigma2FoFcSpin.set_numeric(True);
        s.sigmaFoFcLabel = gtk.Label("FoFc contour (sigma): ");
        adj = gtk.Adjustment(cfg.getfloat("program", "FoFc"),
                             0.00, 50.00, 0.1, -0.1);
        s.sigmaFoFcSpin = gtk.SpinButton(adj, digits=2);
        s.sigmaFoFcSpin.set_numeric(True);
        
        for i, widget in enumerate( (s.sigma2FoFcLabel, s.sigma2FoFcSpin,
                                     s.sigmaFoFcLabel,  s.sigmaFoFcSpin) ):
            s.Tab1.attach(widget, i, i+1, 0, 1, xpadding=5, ypadding=5);

        s.cootWinWidthLabel = gtk.Label("Coot window width: ");
        adj = gtk.Adjustment(cfg.getint("program", "coot-width"),
                             100, 2000.00, 10, -10);
        s.cootWinWidthSpin = gtk.SpinButton(adj, digits=0);
        s.cootWinWidthSpin.set_numeric(True);
        s.cootWinHeightLabel = gtk.Label("Coot window height: ");
        adj = gtk.Adjustment(cfg.getint("program", "coot-height"),
                             100, 2000.00, 10, -10);
        s.cootWinHeightSpin = gtk.SpinButton(adj, digits=0);
        s.cootWinHeightSpin.set_numeric(True);
        
        for i, widget in enumerate( (s.cootWinWidthLabel,s.cootWinWidthSpin,
                                     s.cootWinHeightLabel,s.cootWinHeightSpin)):
            s.Tab1.attach(widget, i, i+1, 1, 2, xpadding=5, ypadding=5);

        s.cootLocLabel = gtk.Label("Coot binary location: ");
        s.cootLocEntry = gtk.Entry();
        s.cootLocEntry.set_max_length(0);
        s.cootLocEntry.set_editable(True);
        s.cootLocEntry.set_property("xalign", 0.0);
        s.cootLocEntry.set_text(cfg.get("program", "cootExec"));
        s.cootLocButton = gtk.Button(stock=gtk.STOCK_OPEN);
        s.cootLocButton.connect("clicked", s._CootLocCB);
        s.Tab1.attach(s.cootLocLabel, 0, 1, 2, 3, xpadding=5, ypadding=5);
        s.Tab1.attach(s.cootLocEntry, 1, 3, 2, 3, xpadding=5, ypadding=5);
        s.Tab1.attach(s.cootLocButton, 3, 4, 2, 3, xpadding=5, ypadding=5);

        s.pdbDirLabel = gtk.Label("PDB mirror directory: ");
        s.pdbDirEntry = gtk.Entry();
        s.pdbDirEntry.set_max_length(0);
        s.pdbDirEntry.set_editable(False);
        s.pdbDirEntry.set_property("xalign", 0.0);
        s.pdbDirEntry.set_text(cfg.get("program", "pdbDir"));
        s.pdbDirButton = gtk.Button(stock=gtk.STOCK_OPEN);
        s.pdbDirButton.connect("clicked", s._PDBDirLocCB);
        s.Tab1.attach(s.pdbDirLabel, 0, 1, 3, 4, xpadding=5, ypadding=5);
        s.Tab1.attach(s.pdbDirEntry, 1, 3, 3, 4, xpadding=5, ypadding=5);
        s.Tab1.attach(s.pdbDirButton, 3, 4, 3, 4, xpadding=5, ypadding=5);
        
        s.tempDirLabel = gtk.Label("Temporary files directory: ");
        s.tempDirEntry = gtk.Entry();
        s.tempDirEntry.set_max_length(0);
        s.tempDirEntry.set_editable(False);
        s.tempDirEntry.set_property("xalign", 0.0);
        s.tempDirEntry.set_text(cfg.get("program", "tempDir"));
        s.tempDirButton = gtk.Button(stock=gtk.STOCK_OPEN);
        s.tempDirButton.connect("clicked", s._TempDirLocCB);
        s.Tab1.attach(s.tempDirLabel, 0, 1, 4, 5, xpadding=5, ypadding=5);
        s.Tab1.attach(s.tempDirEntry, 1, 3, 4, 5, xpadding=5, ypadding=5);
        s.Tab1.attach(s.tempDirButton, 3, 4, 4, 5, xpadding=5, ypadding=5);

        s.ustampLabel = gtk.Label("User ID (abbreviation): ");
        s.ustampEntry = gtk.Entry();
        s.ustampEntry.set_max_length(0);
        s.ustampEntry.set_editable(True);
        s.ustampEntry.set_text(cfg.get("program", "ustamp"));
        s.Tab1.attach(s.ustampLabel, 0, 1, 5, 6, xpadding=5, ypadding=5);
        s.Tab1.attach(s.ustampEntry, 1, 2, 5, 6, xpadding=5, ypadding=5);
        
        s.cootScrHost = gtk.Label("Coot scripting language: ");
        s.cootScrPyBut= gtk.RadioButton(group=None, label="Python",
                                        use_underline=False);
        # these two buttons define a group.
        s.cootScrScmBut= gtk.RadioButton(group=s.cootScrPyBut, label="Scheme",
                                        use_underline=False);
        # line the two radiobuttons in a column, a little bit of waste of
        # space, but it looks better.
        s.Tab1.attach(s.cootScrHost, 0, 1, 6, 7, xpadding=5, ypadding=5);
        s.Tab1.attach(s.cootScrPyBut, 1, 2, 6, 7, xpadding=5, ypadding=5);
        s.Tab1.attach(s.cootScrScmBut, 1, 2, 7, 8, xpadding=5, ypadding=0);
        # choose the right button to be active.
        active = s.cootScrPyBut if cfg.get("program", "scripting-host")=="python"\
                 else s.cootScrScmBut;
        active.set_active(True);
        
        s.cfgLocLabel = gtk.Label("Configuration file: ");
        s.cfgFNLabel = gtk.Label(s.dysConfig.GetCurrentConfigFullFileName());
        # User should be able to copy and paste this path...
        s.cfgFNLabel.set_selectable(True);
        s.Tab1.attach(s.cfgLocLabel, 0, 1, 8, 9, xpadding=5, ypadding=5);
        s.Tab1.attach(s.cfgFNLabel, 1, 4, 8, 9, xpadding=5, ypadding=5);
        
        s.__MakeToolTips();
        
        # make all labels in this window left aligned. Default was centered.
        for label in (s.sigma2FoFcLabel, s.sigmaFoFcLabel, s.cootWinWidthLabel,
                      s.cootWinHeightLabel, s.cootLocLabel, s.pdbDirLabel,
                      s.tempDirLabel, s.ustampLabel, s.cfgLocLabel,
                      s.cootScrHost, s.cfgFNLabel):
            label.set_alignment(xalign=0.0, yalign=0.5);
        s.mainBox.pack_start(s.Tab1, False, False, padding=5);
        
        s.commitBox = gtk.HBox(False, 0);
        s.okButton = gtk.Button(stock=gtk.STOCK_OK);
        s.cancelButton = gtk.Button(stock=gtk.STOCK_CANCEL);
        s.okButton.connect("clicked", s._CommitCB, True);
        s.cancelButton.connect("clicked", s._CommitCB, False);
        s.commitBox.pack_start(s.cancelButton, True, False);
        s.commitBox.pack_start(s.okButton, True, False);
        s.mainBox.pack_start(s.commitBox, padding=5);

    def __MakeToolTips(s):
        "Make the little mouse over texts."
        toolTips = (
            (s.sigma2FoFcSpin, CONTOUR_2FOFC_TTEXT),
            (s.sigmaFoFcSpin, CONTOUR_FOFC_TTEXT),
            (s.cootWinWidthSpin, COOT_WIN_WIDTH_TTEXT),
            (s.cootWinHeightSpin, COOT_WIN_HEIGHT_TTEXT),
            (s.cootLocEntry, COOT_BINARY_TTEXT),
            (s.pdbDirEntry, PDB_MIRROR_DIR_TTEXT),
            (s.tempDirEntry, TEMP_FILES_DIR_TTEXT),
            (s.ustampEntry, USER_ID_TTEXT),
            (s.cootScrScmBut, SCRIPT_SCHEME_TTEXT),
            (s.cootScrPyBut, SCRIPT_PYTHON_TTEXT),
            (s.cfgFNLabel, CONFIG_FILENAME_TTEXT));
        for widget, text in toolTips:
            widget.set_tooltip_text(text);
    
    def _QuitCB(s, widget, data=None):
        """Callback for window delete/destroy events. Treat like Cancel."""
        s._CommitCB(widget, data=False);
        
    def _CommitCB(s, widget, data=False):
        """Callback for cancel/ok button pressed & window destroy.
        """
        cfg = s.dysConfig.GetConfig();
        if( data ):
            # we pressed the OK button, need to store strings in the config.
            cfg.set("program", "2FoFc", str(s.sigma2FoFcSpin.get_value()));
            cfg.set("program", "FoFc", str(s.sigmaFoFcSpin.get_value()));
            cfg.set("program", "coot-width",
                    str(int(s.cootWinWidthSpin.get_value())));
            cfg.set("program", "coot-height",
                    str(int(s.cootWinHeightSpin.get_value())));
            cfg.set("program", "cootExec", s.cootLocEntry.get_text());
            cfg.set("program", "pdbDir", s.pdbDirEntry.get_text());
            cfg.set("program", "tempDir", s.tempDirEntry.get_text());
            cfg.set("program", "ustamp", s.ustampEntry.get_text());
            scriptLang = "python" if s.cootScrPyBut.get_active() else "scheme";
            cfg.set("program", "scripting-host", scriptLang);
            s.dysConfig.WriteConfiguration();
        else:
            # Cancel button pressed. Restore to the previous default values.
            s.sigma2FoFcSpin.set_value(cfg.getfloat("program", "2FoFc"))
            s.sigmaFoFcSpin.set_value(cfg.getfloat("program", "FoFc"))
            s.cootWinWidthSpin.set_value(cfg.getint("program", "coot-width"));
            s.cootWinHeightSpin.set_value(cfg.getint("program", "coot-height"));
            s.cootLocEntry.set_text(cfg.get("program", "cootExec"));
            s.pdbDirEntry.set_text(cfg.get("program", "pdbDir"));
            s.tempDirEntry.set_text(cfg.get("program", "tempDir"));
            s.ustampEntry.set_text(cfg.get("program", "ustamp"));
            scriptLang = cfg.get("program", "scripting-host");
            active = s.cootScrPyBut if scriptLang=="python" else\
                     s.cootScrScmBut;
            active.set_active(True);
        s.hide();
        
    def _CootLocCB(s, widget, data=None):
        "Callback function for coot executable file location."
        cfg = s.dysConfig.GetConfig();
        fn = HandleFileChooser(COOT_FILE_SEL_WINDOW_TITLE,
                               gtk.FILE_CHOOSER_ACTION_OPEN,
                               suggestedFN=cfg.get("program", "cootExec"));
        if( fn!=None ):
            if(   not os.path.exists(fn) ):
                OkInfoDialog(FILE_NOT_EXIST_TEXT);
                return ;
            elif( not os.access(fn, os.X_OK) ):
                OkInfoDialog(FILE_NOT_EXECTUTABLE_TEXT);
                return ;
            # integrity check passed.
            s.cootLocEntry.set_text(fn);

    def _PDBDirLocCB(s, widget, data=None):
        cfg = s.dysConfig.GetConfig();
        fn = HandleFileChooser(PDB_DIR_SEL_WINDOW_TITLE,
                               gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER,
                               suggestedFN=cfg.get("program", "pdbDir"));
        if( fn!=None ):
            if(   not os.path.isdir(fn) ):
                # This should *never* happen...
                OkInfoDialog(NOT_A_DIRECTORY_TEXT);
                return ;
            elif( not os.access(fn, os.X_OK) ):
                OkInfoDialog(CANNOT_CHG2DIRECTORY_TEXT);
            # integrity check passed.
            s.pdbDirEntry.set_text(fn);
    
    def _TempDirLocCB(s, widget, data=None):
        cfg = s.dysConfig.GetConfig();
        fn = HandleFileChooser(PDB_DIR_SEL_WINDOW_TITLE,
                               gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER,
                               suggestedFN=cfg.get("program", "tempDir"));
        if( fn!=None ):
            if(   not os.path.isdir(fn) ):
                # This should *never* happen...
                OkInfoDialog(NOT_A_DIRECTORY_TEXT);
                return ;
            elif( not os.access(fn, os.X_OK) ):
                OkInfoDialog(CANNOT_CHG2DIRECTORY_TEXT);
                return ;
            elif( not os.access(fn, os.W_OK) ):
                OkInfoDialog(NO_WRITE_PERMS_TEXT);
                return ;
                
            # integrity check passed.
            s.tempDirEntry.set_text(fn);
            
    def _PymolLocCB(s, widget, data=None):
        fn = HandleFileChooser(PYMOL_FILE_SEL_WINDOW_TITLE,
                               gtk.FILE_CHOOSER_ACTION_OPEN);
        if( fn!=None ):
            s.pymolExecFFN = fn;
            s.pymolLocEntry.set_text(s.pymolExecFFN);
    
            
class CImportEDSnPDB(gtk.Dialog):
    """Import widget.

    Upon run() invocation, this widget returns 'gtk.RESPONSE_OK' if the OK
    button was pressed. 
    """
    def __init__(s):
        gtk.Dialog.__init__(s, title=IMPORTEDSPDB_WINDOW_TITLE, 
                   parent=None, flags=gtk.DIALOG_MODAL,
                   buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                            gtk.STOCK_OK, gtk.RESPONSE_OK));
        s._PdbCode = None;
        s.__BuildWindow();
        # This call here alone is futile. The entry widget needs to be enabled
        # for default actions, that is: entry.set_activates_default(True), see
        # below in __BuildWindow.
        s.set_default_response(gtk.RESPONSE_OK);
        s.show_all();

    def __BuildWindow(s):
        # There is already a vbox in this class.
        # Table row and columns are adjusted to their max element.
        s.Tab = gtk.Table(1, 2, homogeneous=False);
        s.importRemoteLabel = gtk.Label("PDB 4 letter code: ");
        s.importRemoteEntry = gtk.Entry();
        s.importRemoteEntry.set_max_length(0);
        s.importRemoteEntry.set_editable(True);
        s.importRemoteEntry.set_property("xalign", 0.0);
        s.importRemoteEntry.set_tooltip_text(EDS_PDB_ENTRY_TTEXT);
        # Allow default actions for this input widget.
        s.importRemoteEntry.set_activates_default(True);
        s.Tab.attach(s.importRemoteLabel, 0, 1, 0, 1, xpadding=5, ypadding=5);
        s.Tab.attach(s.importRemoteEntry, 1, 2, 0, 1, xpadding=5, ypadding=5);
        s.vbox.pack_start(s.Tab, False, False, padding=5);

    def GetPDBCode(s):
        """Get the PDB code entered in the dialog box.
        @return pdbCode  Either the 4 letter code or @c None to indicate that
         the process was canceled.
        """
        return s.importRemoteEntry.get_text();
    

def YesNoInfoDialog(msgText):
    """Pop up modal yes/no info dialog with text 'msgText'.
    Return True if 'Yes' was chosen, else False.
    """
    dlg  = gtk.MessageDialog(buttons=gtk.BUTTONS_YES_NO,
                             type=gtk.MESSAGE_WARNING, flags=gtk.DIALOG_MODAL);
    dlg.set_property("text", msgText);
    dlg.set_title("Information")
    retVal = dlg.run();
    dlg.destroy();
    return (retVal not in  (gtk.RESPONSE_NONE, gtk.RESPONSE_NO));

def OkInfoDialog(msgText):
    """Pop up modal info dialog with text 'msgText'.
    """
    dlg  = gtk.MessageDialog(buttons=gtk.BUTTONS_OK,
                             type=gtk.MESSAGE_WARNING, flags=gtk.DIALOG_MODAL);
    dlg.set_property("text", msgText);
    dlg.set_title("Information")
    retVal = dlg.run();
    dlg.destroy();

def SimpleOKDialog(msgText):
    """Pop up info dialog with text 'msgText'.
    """
    dlg  = gtk.MessageDialog(buttons=gtk.BUTTONS_OK)
    dlg.set_property("text", msgText);
    dlg.set_title("Information")
    retVal = dlg.run();
    dlg.destroy();

def HandleFileChooser(title, action, suggestedFN=""):
    """Helper function for file reading/writing and directory choosing dialog.
    - title: Window title.
    - action: One of gtk.FILE_CHOOSER_ACTION_OPEN,
      gtk.FILE_CHOOSER_ACTION_SAVE, gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER or
      gtk.FILE_CHOOSER_ACTION_CREATE_FOLDER
    - suggested: Suggested file/directory name.
    return None if nothing chosen, else the name of the chosen file.
    """
    if(   action==gtk.FILE_CHOOSER_ACTION_OPEN ):
        buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                   gtk.STOCK_OPEN, gtk.RESPONSE_OK);
    elif( action==gtk.FILE_CHOOSER_ACTION_SAVE ):
        buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                   gtk.STOCK_SAVE_AS, gtk.RESPONSE_OK);
    elif( action==gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER ):
        buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                   gtk.STOCK_OPEN, gtk.RESPONSE_OK);
    dlg = gtk.FileChooserDialog(title, None, action,buttons=buttons);
    dlg.set_default_response(gtk.RESPONSE_OK);
    dlg.set_select_multiple(False);
    if( suggestedFN!="" and action==gtk.FILE_CHOOSER_ACTION_SAVE ):
        dlg.set_current_name(suggestedFN);
    elif( suggestedFN!="" and action==gtk.FILE_CHOOSER_ACTION_OPEN ):
        # This is mainly to switch to the directory the file name belongs to.
        dlg.set_filename(suggestedFN);
    response = dlg.run();
    chosenName = None;
    if( response==gtk.RESPONSE_OK ):
        chosenName = dlg.get_filename();
    dlg.hide();
    dlg.destroy();
    return chosenName;


if( __name__=="__main__" ):
    ap=argparse.ArgumentParser(description="""This program displays the table
    of EDS ligand data and allows to edit comments. Coot is started for a
    selected line.""", epilog="""When launched without arguments, eg. from a
    GUI, the program displays an empty table. Data can be loaded from a file
    menu.""");
    ap.add_argument("-v", "--version",
                    help="Print version information and exit.",
                    action="version", version=__version__);
    ap.add_argument("--latest", dest="loadLatest",
                    help="""Just provide the base name without version number,
                    and the program will search for the highest version number
                    of the table and load it. The succeeding version number
                    will be used to save the file. It is granted that this
                    version number file does not yet exist. """, 
                    action="store_true");
    ap.add_argument("--config", type=str, 
                    help="""File name of alternative config file to load. If
                    it does not yet exist, it will be created and filled with
                    default values. This file can then be edited.""",
                    default="");
    ap.add_argument("--read", type=str, help="Input file with EDS table.",
                    dest="edstable", metavar="FILE", default=EMPTYFN);
    ap.add_argument("--show-hq-rscc", dest="goodLigands", action="store_true",
                    help="""Instead of viewing the table of ligands with low
                    RSCC (<%3.2f) values, show a table with very high RSCC
                    values (>=%3.2f).""" % (RSCCBADTH, RSCCGOODTH));
    ap.add_argument("--sort-rscc", dest="sortRSCC", action="store_true",
                    help="""Group sort the table by RSCC value only, instead
                    of combined score S.""");
    
    todo=ap.parse_args();
    edsInFN, edsOutFN    = GetVersionedInOutFileNames(todo.edstable);
    edsTableFN  = edsInFN if todo.loadLatest else todo.edstable;
    
    app = CGtkDysplayer(edsTableFN, todo.config, sortS=not todo.sortRSCC,
                        badLigands=not todo.goodLigands);
    gtk.main();




# IO of ligand table.
#
# Author       : Chris X. Weichenberger
# Created      : 2012-02-24
# License
# Twilight is a GUI for displaying ligands incompatible with electron density.
#    Copyright (C) 2012  Christian X. Weichenberger, Edwin Pozharski, and
#    Bernhard Rupp.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Modifications:
# 2012-03-15 (CW) Sorted list of het groups, Ed's get for all bad valid ligands.
# 2012-06-14 (CW) Removed truncation of Jrnl column to have full journal name.
# 2012-07-27 (CW) Adding of ligands.
# 2012-10-15 (CW) Use of deposit date instead of release date.
# 2012-10-14 (CW) Rewrote meaning of bad ligands. Good and bad ones.
# 2012-10-16 (CW) Complex sorting for an arbitrary column.
#

import csv

from misc import CalcCombinedScore, VALID_BAD, VALID_GOOD, VALID_ADDED,\
     VALID_NEITHER, RSCCBADTH, RSCCGOODTH;
from files import Output;

# Number of columns in this file.
NR_COLUMNS = 22;

class CEDSTableData(list):
    # Here we have some class variables.
    PDBIDIDX, LIGNMIDX, RESNRIDX, RSCCIDX, OWABIDX, MOLWTIDX, CHARGEIDX, \
    HEAVYIDX, HYDROGIDX, ATOMSIDX, DATIDX, RELIDX, RWORKIDX, RFREEIDX, \
    ROBSIDX, SWIDX, JRNLIDX, PMIDIDX, CLASSIDX, VALIDIDX, SCOREIDX, \
    COMMENTIDX =\
    range(NR_COLUMNS);
    ColumnTypes = (str   , str , str , float , float , float , int   , 
                   int   , int , int , str   , float , float , float ,
                   float , str , str , str   , str   , str   , float ,
                   str    );

    def __init__(s):
        s._Header     = ();
        # Default sorting: by combined score and descending.
        s.SetSortingScheme(s.SCOREIDX, False);
        # Default ligands on display: bad
        s.SetFocusOnBad()
    
    def _LigandInFocus(s, rscc):
        if( s._FocusIsBad ):
            return rscc<RSCCBADTH;
        else:
            return rscc>=RSCCGOODTH;
            
    def IsValidTableFile(s, filename):
        """Checks if the file looks like an EDS table."""
        fh     = open(filename);
        header = fh.readline();
        fh.close();
        if( header[:5]=="PDBID" ):
            # looks good, let's check the number of columns...
            header = header.split("\t");
            if( len(header)==NR_COLUMNS ):
                return True;
        return False;
        
    def ReadTable(s, filename):
        """Opens and reads an EDS table.
        @param filename  Full path file name.
        """
        Output("Reading table ``%s''..." % filename);
        fh     = open(filename);
        reader = csv.reader(fh, delimiter="\t");
        s._Header = reader.next();
        s._All = [l for l in reader];
        fh.close();
        s._ComplexSort();
        s._ReIndex();

    def SetFocusOnBad(s):
        """Gather bad ligands. Call this prior to loading a file!"""
        s._FocusIsBad = True;
    
    def SetFocusOnGood(s):
        """Gather good ligands. Call this prior to loading a file!"""
        s._FocusIsBad = False;
    
    def ToggleFocus(s):
        """Utility function to invert the outcome of the Get*Ligands methods.
        """
        s._FocusIsBad = not s._FocusIsBad;
    
    def SetSortingScheme(s, sortCol, ascending):
        """Set the sorting column and order.
        @param sortCol  Column index.
        @param ascending @c True or @c False.
        @return @c False if sortCol was incorrect, @c True, if ok.
        """
        if( sortCol<0 or sortCol>=s.COMMENTIDX ):
            return False;
        s._SortColIdx = sortCol;
        s._SortAsc    = ascending;
        
    def IsFocusOnBad(s):
        """What type of ligands do we display?
        @return @c True, bad ones. @c False, good ones.
        """
        return s._FocusIsBad;
    
    def IsManuallyImported(s, pos):
        return s[pos][s.VALIDIDX]==VALID_ADDED;
    
    def GetHeader(s):
        return s._Header;

    def GetPDBID(s, pos):
        return s[pos][s.PDBIDIDX];
    
    def GetResNr(s, pos):
        return s[pos][s.RESNRIDX];

    def GetLigNm(s, pos):
        return s[pos][s.LIGNMIDX];
    
    # formerly: GetAllValidLigands
    def GetAllDisplayLigands(s, pos):
        """
        Get all ligands which should be displayed in the table for this PDB
        entry.
        """ 
        pdbID = s.GetPDBID(pos);
        wanted= [VALID_ADDED, ];
        wanted.append(VALID_BAD if s._FocusIsBad else VALID_GOOD);
        res   = [item[s.RESNRIDX] for item in s._All
                 if item[s.PDBIDIDX]==pdbID and item[s.VALIDIDX] in wanted];
        res.sort();
        return res;

    # formerly: GetAllBadRSCCLigands
    def GetAllRSCCFocusLigands(s, pos):
        """Get all ligands which are currently ligands of desired quality.
        This depends on whether we  want to display ligands with bad RSCC
        values or with good RSCC values.
        """
        pdbID = s.GetPDBID(pos);
        res   = [item[s.RESNRIDX] for item in s._All
                 if item[s.PDBIDIDX]==pdbID and
                 s._LigandInFocus(float(item[s.RSCCIDX]))];
        res.sort();
        return res;
    
    # formerly: GetAllValidBadRSCCLigands
    def GetAllDisplayRSCCFocusLigands(s, pos, fsorted=True):
        """Get all ligands that should be displayed in the table and meet the
        desired RSCC criterion (either good or bad values, depending of the
        focus).
        """
        pdbID  = s.GetPDBID(pos);
        wanted = [VALID_ADDED, ];
        wanted.append(VALID_BAD if s._FocusIsBad else VALID_GOOD);
        res   = [item[s.RESNRIDX] for item in s._All
                 if item[s.PDBIDIDX]==pdbID and
                 s._LigandInFocus(float(item[s.RSCCIDX])) and
                 item[s.VALIDIDX] in wanted];
        if fsorted:
            res.sort();
        return res;
        
    def GetPreviousInPDB(s, pos):
        pdbID = s.GetPDBID(pos);
        for i in range(pos-1,-1,-1):
            if( s[i][s.PDBIDIDX]==pdbID ):
                return i;
        return pos;

    def ConvertRow(s, i):
        """Convert the text in each row to its natural data types."""
        return [dType(col) for dType, col in zip(s.ColumnTypes, s[i])];
        
    def WriteTSV(s, filename, verbose=False):
        """Save file to given file name.
        @retun  @c True, if everything went ok. @c False if an error during
        file opening occurred.
        """
        if( verbose ):
            Output("\rWriting table ``%s''..." % filename);
        for idx, entry in zip(s._Idx, s):
            s._All[idx][s.COMMENTIDX] = entry[s.COMMENTIDX];
        try:
            fh      = open(filename, "w");
            content = "\n".join(["\t".join(line)
                                 for line in [s.GetHeader(),]+s._All]);
            fh.write(content);
            fh.close();
            return True;
        except IOError:
            return False;

    def _ComplexSort(s):
        """Sorts s._All in a meaningful way.
        That is, PDB entries are grouped. Within each PDB group, those rows
        with a valid column are sorted by Score and ranked top. After them,
        columns with valid code VALID_NEITHER, sorted by Score, are added to
        the group. This guarantees that the highest Score is on top of the whole
        group. All groups are again sorted by the Score of its first row,
        therefore ranking the worst ligands per group (ie. PDB entry) on top.
        """
        desc     = lambda x,y: -(x>y);
        asc      = lambda x,y: -(x<y);
        sortFunc = asc if s._SortAsc else desc;
        # sorting within a group, this is by Score.
        perGroup = lambda x,y,i=s._SortColIdx: sortFunc(float(x[i]),float(y[i]));
        # sorting of the groups, this is by the Score of the first row.
        wholeTab = lambda x,y,i=s._SortColIdx: sortFunc(float(x[1][0][i]), 
                                        float(y[1][0][i]));
        # Partition data set into a valid and bad VALID_BAD, a valid and good
        # VALID_GOOD, an added VALID_ADDED, and not valid (ie. neither good
        # nor bad, unexciting) VALID_NEITHER column set.
        pdbBuckets = {VALID_ADDED: {}, VALID_BAD: {}, VALID_GOOD: {},
                      VALID_NEITHER: {}};
        for row in s._All:
            validCode = row[s.VALIDIDX];
            pdbBuckets[validCode].setdefault(row[s.PDBIDIDX], []).append(row[:]);
        # sort the two data sets by Score.
        for validCode, target in pdbBuckets.items():
            for rows in target.values():
                rows.sort(perGroup);
        
        # Merge data sets again.
        # On the very top we put the manually added PDB entries.
        # Below, we place those groups with the highest Score for the valid
        # columns (per group!). Append not valid rows after all valid rows.
        inFocus, outFocus = [VALID_BAD, VALID_GOOD] if s._FocusIsBad else \
                            [VALID_GOOD, VALID_BAD];
        allItems = pdbBuckets[inFocus].items();
        allItems.sort(wholeTab);
        newList = [];
        for rows in pdbBuckets[VALID_ADDED].values():
            newList.extend(rows);
        seen    = set();
        for key, rows in allItems:
            # first the valid rows in focus, worst on top of this group.
            newList.extend(rows);
            # then add those which are currently not in the focus
            newList.extend(pdbBuckets[outFocus].get(key, []));
            # and then append any not valid rows.
            newList.extend(pdbBuckets[VALID_NEITHER].get(key, []));
            seen.add(key);
        # finally complete the list with all groups that did not have a valid
        # row at all.
        for key, rows in pdbBuckets[VALID_NEITHER].items():
            if( key not in seen ):
                newList.extend(rows);
        # replace the original table with this sorted table.
        s._All = newList;

    def _ReIndex(s):
        """Create internal indices.
        This is needed when data have been loaded or freshly added. 
        """
        wanted= [VALID_ADDED,];
        wanted.append(VALID_BAD if s._FocusIsBad else VALID_GOOD);
        s._Idx = [n for n,l in enumerate(s._All) if l[s.VALIDIDX] in wanted];
        s[:] = [l[:] for l in s._All if l[s.VALIDIDX] in wanted];
        
    def AddLigand(s, pdbID, pdbFileObj, ligandObj):
        """Adds a ligand to the top of this list.

        @param pdbID  PDB four letter code.
        @param pdbFileObj  @c CPDBFile object.
        @param ligandObj  @c CLigand object.
        """
        lig = ligandObj;
        pdb = pdbFileObj;
        score = "%4.3f" % CalcCombinedScore(lig.GetRSCC(),pdb.GetResolution());
        elem = [pdbID,
                lig.Get3LetterName(),
                lig.GetResNr(),
                "%3.3f" % lig.GetRSCC(),
                "%4.2f" % lig.GetOWAB(),
                "%5.1f" % lig.GetMolWt(),
                "%2d" % lig.GetCharge(),
                "%3d" % lig.GetNrHeavyAtoms(),
                "%3d" % lig.GetNrHydrogenAtoms(),
                "%3d" % lig.GetNrAtoms(),
                pdb.GetDepositDate(),
                "%3.2f" % pdb.GetResolution(),
                "%4.3f" % pdb.GetRWork(),
                "%4.3f" % pdb.GetRFree(),
                "%4.3f" % pdb.GetRObs(),
                pdb.GetRefinementSoftware(),
                pdb.GetJournal(),
                pdb.GetPubMedID(),
                lig.GetClassification(),
                VALID_ADDED,
                score,
                ""];
        # Re-indexing is guearanteed to work when ligand data has been entered
        # at the top of the list.
        s._All.insert(0, elem);
        s._ReIndex();

    def RemoveLigand(s, pdbID, resNr):
        # probably we need this with an index instead of pdb/rnr...
        """Remove a ligand from the List and the file.
        @param pdbID  PDB four letter code.
        @param resNr  Residue number with chain name and insertion code.
        """
        for i, l in enumerate(s._All):
            if( l[s.PDBIDIDX]==pdbID and l[s.RESNRIDX]==resNr ):
                del s._All[i];
                break ;
        s._ReIndex();

if( __name__=="__main__" ):
    obj=CEDSTableData();
    obj.ReadTable("ligands-2012-01-15.tsv");
    for line in obj[:20]:
        print line;



# Some basic file io.
#
# Author       : Chris X. Weichenberger
# Created      : 2011-07-13
# License
# Twilight is a GUI for displaying ligands incompatible with electron density.
#    Copyright (C) 2012  Christian X. Weichenberger, Edwin Pozharski, and
#    Bernhard Rupp.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Modifications:
#  2012-05-14 (CW) ReadTextFile treats non-newline terminated lines correctly.
#  2012-08-01 (CW) Data directory now dependent on the location of this file.
#  2012-08-02 (CW) Fixed table separator in CTable, can be supplied now.
#  2012-10-18 (CW) Some more constants.

import os
import sys
import csv
import string
import platform

# Directory where installation resides.
INSTALLDIR = os.path.abspath(os.path.dirname(__file__));
# main directory with all data files.
DATADIR    = os.path.join(INSTALLDIR, "data");
# Directory where EDS data files have been downloaded to. Incremental update.
EDSDIR     = os.path.join(DATADIR, "EDSDB");
# A version with release date and software string added.
PDBLISTFN  = os.path.join(DATADIR, "PDBentries-4ligands-2012-01-15-d.csv");
# Information about refinement program which is not available in PDB adv query.
REFINEPROGS= os.path.join(DATADIR, "refineprogs-2012-01-15.tsv");
# Periodic table of the elements.
ELEMENTSFN = os.path.join(DATADIR, "list-of-elements.tsv");
# Empty table template file name.
EMPTYFN    = os.path.join(DATADIR, "template-22-cols.tsv");
# File name for new file from scratch.
UNTITLEDFN = "Untitled.tsv";
# Redirected output file name of combine.py, also input file for postfilter.py
COMBINEDFN = "combine-2012-01-15-d.tsv";
# Patch for conversion Release Date -> Deposit Date
DEPOSITFN  = os.path.join(DATADIR,"rel-dep-date-2012-10-15-for-2012-01-15.tsv");
LPETIDESFN = os.path.join(DATADIR, "L-peptides.csv");
BUFFERNMFN = os.path.join(DATADIR, "buffermolecules.csv");
THREE2ONE  = { "GLY": "G", "ALA": "A", "VAL": "V", "SER": "S", "THR": "T",
               "CYS": "C", "PRO": "P", "LEU": "L",    "ILE": "I", "ASP": "D",
               "ASN": "N", "HIS": "H", "PHE": "F", "TYR": "Y", "TRP": "W",
               "MET": "M",    "GLU": "E", "GLN": "Q", "LYS": "K", "ARG": "R" }
STDAMINOS  = THREE2ONE.keys();
NONLIGANDS = STDAMINOS + ["MSE", "HOH", "DOD"];

STDBASES   = [ "  A", "  C", "  G", "  T", "  U"];
DBASES     = [ " DA", " DC", " DG", " DT", " DU"];

# This should be used wherever an expected float value is not available.
NOVALIDFLOATSTR = "-66.6";
# Invalid date, eg. if PDB entry has become obsolote or parsing error.
INVALIDPDBDATE  = "01-JAN-60";
# The same date, but in the format as it is stored in the table.
INVALIDTABDATE  = "2060-01-01";
# Indices of column in data files. As data is appended in several steps, this
# should remain a constant, as long as they are present in the respective
# file. 
PDBIDI    = 0;
RESNMI    = 1;
RESNRI    = 2;
RSCCI     = 3;
NRATMSI   = 7;
DATEI     = 10;
RESOLI    = 11;
CLASSI    = 18;
VALIDCOLI = 19;

# Types of valid molecules
# Valid by filter criteria and bad RSCC.
VALID_BAD     = "Y";
# Valid by filter criteria and good RSCC.
VALID_GOOD    = "G";
# Manually added.
VALID_ADDED   = "A";
# None of the above.
VALID_NEITHER = "N"

# RSCC threshold used in the postfilter.py program.
RSCCBADTH     = 0.6;
RSCCGOODTH    = 0.95;       


def ReadTextFile(ffn):
    """Read a text file and return its content. Cut newlines at the end
    * return: List of Strings or None if file could not be found.
    """
    try:
        fh    = open(ffn);
        lines = [x[:-1] if x[-1]=="\n" else x for x in fh.readlines()];
        fh.close();
        return lines;
    except IOError:
        sys.stderr.write("Could not open file ``%s''\n" % ffn);
        return None

def CalcCombinedScore(rscc, resol):
    """Compute the combined score from RSCC and experimental resolution.
    @param rscc  RSCC value.
    @param resol  Resolution of X-ray structure.
    """
    return 2.0/(abs(float(rscc))/RSCCBADTH + float(resol)/1.3)

class CTable(list):
    def __init__(s, ffn, ints={"key": "%d"}, floats={"key": "%3.2f"}, sep="\t"):
        """Reads in a table and converts rows to their respective data types.
        Keys are the column headers!
        s.FFN: Full file name.
        @param sep  Separator for columns.
        """
        s._Ints  = ints;
        s._Floats= floats;

        list.__init__(s);
        s._ReadTable(ffn, sep);
        
    def _ReadTable(s, ffn, sep):
        """Read a tabular file"""
        s.FFN = ffn;
        fh = open(ffn);
        reader = csv.reader(fh, delimiter=sep)
        s._Headers = reader.next();
        for line in reader:
            di = {};
            for k, v in zip(s._Headers, line):
                if( k in s._Ints):   v = int(v);
                if( k in s._Floats):
                    if( v!="" ):
                        v = float(v);
                    else:
                        v = float(NOVALIDFLOATSTR);
                di[k] = v;
            s.append(di);
    
    def GetFileName(s):
        return s.FFN;

    def AsTable(s, remove=(), add=()):
        order = [x for x in s._Headers if x not in remove] + \
                [x for x in add];
        tab = [ [elem[k] for k in order] for elem in s ];
        return tab;
            
    def Write(s, fh=sys.stdout, remove=(), add=(),
              header=True, sortCol="", sep="\t"):
        """Write the table.
        @param fh  Output file handle.
        @param remove  List of column names to remove in the table.
        @param add  List of column names to add to the output.
        @param header  True to output header line.
        @param sortCol  Name of column to sort before outputting.
        @param sep  Column separator in output file.
        """
        order = [x for x in s._Headers if x not in remove] + \
                [x for x in add];

        if( sortCol!="" ):
            s.sort( lambda a,b: -(a[sortCol]<b[sortCol]) );        
        output= [];
        if( header ):
            fh.write("\t".join(order) + os.linesep );
        for elem in s:
            lst = [];
            for key in order:
                if(   key in s._Ints ):
                    st = s._Ints[key] % elem[key] ;
                elif( key in s._Floats ):
                    st = s._Floats[key] % elem[key];
                else:
                    # string
                    st = elem[key];
                lst.append(st);
            output.append( sep.join(lst) );
        fh.write( "\n".join(output) );
        if( len(output)>=1 ):
            # string.join does make N-1 concatenations for N entries in a list. 
            fh.write("\n");



# PDB file, coot script file handling.
#
# Author       : Chris X. Weichenberger
# Created      : 2012-02-24
# License
# Twilight is a GUI for displaying ligands incompatible with electron density.
#    Copyright (C) 2012  Christian X. Weichenberger, Edwin Pozharski, and
#    Bernhard Rupp.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Modifications:
# 2012-03-15 (_X) Sorted list of het groups.
# 2012-07-24 (CW) Import of PDB and EDS data directly from EDS.
# 2012-08-01 (CW) Fix for TNT refinement program parsing of R-values.
# 2012-08-02 (CW) Compute the ligands' mol. weight when parsing the PDB file.
# 2012-10-10 (CW) Better discrimination between failures. Loading from PDB and
#  from EDS in order to find out if a PDB entry is missing due to electron
#  density unavailability.
# 2012-10-15 (CW) Use of deposit date instead of release date.

import os
import sys
import gzip
import urllib
import shutil
import tempfile

from misc     import NOVALIDFLOATSTR, NONLIGANDS, INVALIDPDBDATE;
from ligand   import CLigand;
from eds      import CEDSFile, EDSRESNMIDX, EDSRSCCIDX, EDSOWABIDX;

from files import Debug;

# Error message constants.
# PDB file retrieved, everything ok.
PDB_OK                  = 0;
# Could not establish internet connection.
PDB_NO_CONNECTION       = 1;
# Could not find entry/file on PDB server.
PDB_PDBENTRY_NOT_EXISTS = 2;
# Could not find entry/file on EDS server.
PDB_EDSENTRY_NOT_EXISTS = 3;
# Invalid Mapman file from EDS.
PDB_EDS_INVALID_STATLIS = 4;
# No ligand data present in PDB file.
PDB_NO_LIGANDS_FOUND    = 5;

class CPDBFile(list):
    EDSBaseURL= "http://eds.bmc.uu.se/eds/dfs/%s/%s/";
    PDBBaseURL= "http://www.rcsb.org/pdb/files/%s.pdb";
    def __init__(s):
        "Handle PDB/EDS data. dataDir may point to a PDB archive mirror." 
        s._PDBDir         = "";
        s._EDSDir         = "";
        s._CootDir        = "";
        s._CootExec       = "";
        s._CurrentPDBID   = "";
        s._ShellScriptExt = ".bat" if sys.platform=="win32" else ".sh";
        # this option should be overridden by a call from the config module.
        s._ScriptingHost = "python";
    
    def _GetLocalPDBFileName(s, pdbID):
        return os.path.join(s._CootDir, pdbID + ".ent");
    
    def _GetMirrorPDBFileName(s, pdbID):
        return os.path.join(s._PDBDir, pdbID[1:3], "pdb%s.ent.gz" % pdbID)

    def _GetLocalMTZFileName(s, pdbID):
        return os.path.join(s._CootDir, pdbID + "_sigmaa.mtz");

    def _GetMirrorEDSStatLisFileName(s, pdbID):
        return os.path.join(s._EDSDir, pdbID + "_stat.lis");
    
    def _GetLocalEDSStatLisFileName(s, pdbID):
        return os.path.join(s._CootDir, pdbID + "_stat.lis");
        
    def _GetLocalCootScriptFileName(s, pdbID):
        ext = ".py" if s._ScriptingHost=="python" else ".scm";
        return os.path.join(s._CootDir, pdbID + ext);

    def _GetLocalShellScriptFileName(s, pdbID):
        return os.path.join(s._CootDir, pdbID + s._ShellScriptExt);
    
    def _GroupLines(s):
        s._Records = {};
        for line in s:
            s._Records.setdefault(line[0:6], []).append(line);
    
    def _DetermineHetGroups(s):
        s._HetGroups = {};
        for line in s._Records.get("HETATM", []):
            rnr = line[21:27];
            rnm = line[17:20];
            if( rnm in NONLIGANDS ): continue 
            if( s._HetGroups.get(rnr, None)==None ):
                chemFormula = "";
                for chemLine in s._Records.get("FORMUL", []):
                    if( chemLine[12:15]==rnm ):
                        chemFormula = chemLine[19:70].strip();
                ligand = CLigand();
                ligand.SetResNr(rnr);
                ligand.Set3LetterName(rnm);
                ligand.SetChemFormula(chemFormula);
                ligand.SetMolWt( ligand.ComputeMolWeight() );
                s._HetGroups[rnr] = ligand;
                
    def _ConvertPDBDate(s, date):
        m2m  = {"JAN": "01", "FEB": "02", "MAR": "03", "APR": "04",
                "MAY": "05", "JUN": "06", "JUL": "07", "AUG": "08",
                "SEP": "09", "OCT": "10", "NOV": "11", "DEC": "12"};
        try:
            dd, mm, yy = date.split("-");
        except ValueError:
            # If the date format does not produce three values, take the
            # invalid date.
            dd, mm, yy = INVALIDPDBDATE.split("-");
        yyprefix = "19" if int(yy)>=60 else "20";
        return "%s-%s-%s" % (yyprefix+yy, m2m.get(mm, "00"), dd);
    
    def SetPDBDir(s, pdbDir):
        """Directory name or empty @c String."""
        # XXX this needs some more work...
        s._PDBDir = pdbDir;

    def SetEDSDir(s, edsDir):
        """Directory name or empty @c String."""
        s._EDSDir  = edsDir;
        
    def SetCootTmpDir(s, cootDir):
        s._CootDir = cootDir;
    
    def SetCootExec(s, cootExec):
        s._CootExec = cootExec;
    
    def SetCootScriptingHost(s, lang):
        s._ScriptingHost = lang;
    
    def GetAllHetGroups(s):
        """
        @return  @c List of @c Tuple (residue number, CLigand object).
        """
        res = [(rnr, ligand) for rnr, ligand in s._HetGroups.items() ];
        res.sort();
        return res;
    
    def GetCenterAtom(s, resNr):
        for line in s:
            if( line[:6]!="HETATM" ): continue ;
            if( line[21:27]==resNr ):
                # atom name of first atom of this group.
                return line[12:16]; 
        return "";

    def GetReleaseDate(s):
        date = INVALIDPDBDATE;
        for line in s._Records.get("REVDAT", []):
            if( line[:11]=="REVDAT   1 " ):
                date = line[13:22];
        return s._ConvertPDBDate(date);
    
    def GetDepositDate(s):
        noHeader = ["HEADER    XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX   %s   9XXX              " % INVALIDPDBDATE,]
        date = s._Records.get("HEADER", noHeader)[0][50:59];
        return s._ConvertPDBDate(date);
        
    def GetResolution(s):
        """Get experimental resolution. Also for non-xray stuff."""
        resol   = NOVALIDFLOATSTR;
        for line in s._Records.get("REMARK", []):
            if( line[:11]=="REMARK   2 " ):
                parts = line.split();
                try:
                    resol = parts[3];
                except IndexError:
                    resol = NOVALIDFLOATSTR;
                try:
                    float(resol);
                except ValueError:
                    resol = NOVALIDFLOATSTR;
        return float(resol);

    def GetR(s):
        def convertR(line):
            try:
                val = float(line.split(":")[1]);
            except:
                val = float(NOVALIDFLOATSTR);
            return val;
        # There are multiple blocks addressing the R values. We need to be
        # sure to parse the right one. It starts with a REMARK   3  FIT TO
        # DATA USED IN REFINEMENT. Once we have parsed it, another block
        # (FIT/AGREEMENT for example) may not be parsed.
        rValues       = [float(NOVALIDFLOATSTR), float(NOVALIDFLOATSTR),
                         float(NOVALIDFLOATSTR)];
        haveRefineSec = False;
        fitTok        = "REMARK   3  FIT";
        fitTNTTok     = "REMARK   3  USI";
        refineTok     = "REMARK   3  FIT TO DATA USED IN REFINEMENT";
        refineTNTTok  = "REMARK   3  USING DATA ABOVE SIGMA CUTOFF.";
        rWrkOrObsTok  = "REMARK   3   R VALUE";
        rFreTok       = "REMARK   3   FREE R VALUE     ";
        for line in s._Records.get("REMARK", []):
            if( line[:15] in (fitTok, fitTNTTok) and haveRefineSec ):
                haveRefineSec = False;
                continue ;
            if( line[:42] in (refineTok, refineTNTTok) ):
                haveRefineSec = True;
            if(   haveRefineSec and line[:20]==rWrkOrObsTok ):
                words = line.split();
                if( "(WORKING" in words and "TEST" in words ):
                    rValues[0] = convertR(line);
                elif( "(WORKING" in words and "TEST" not in words ):
                    rValues[1] = convertR(line);
            elif( haveRefineSec and line[:30]==rFreTok ):
                rValues[2] = convertR(line);
        return rValues;
    
    def GetRObs(s):
        return s.GetR()[0];

    def GetRWork(s):
        return s.GetR()[1];

    def GetRFree(s):
        return s.GetR()[2];

    def GetRefinementSoftware(s):
        refProgs = [];
        for line in s._Records.get("REMARK", []):
            if( line[:26]=="REMARK   3   PROGRAM     :" ):
                progStr = line[27:].strip();
                for programPlusVersion in progStr.split(", "):
                    components = programPlusVersion.split(" ");
                    # this gets rid of version numbers.
                    refProgs.append(components[0]);
        return ",".join(refProgs);

    def GetJournal(s):
        for line in s._Records.get("JRNL  ", []):
            if( line[:16]=="JRNL        REF " ):
                if( line[:34]!="JRNL        REF    TO BE PUBLISHED" ):
                    return line[19:47].strip();
        return "To be published";

    def GetPubMedID(s):
        for line in s._Records.get("JRNL  ", []):
            if( line[:16]=="JRNL        PMID" ):
                try:
                    # I saw one (now already updated) PDB file with missing
                    # PMID, even though the line was present (2w1p).
                    return line.split()[2];
                except IndexError:
                    pass ;
        return "-";

    def FetchPDB(s, pdbID):
        """Fetch PDB file from EDS server or from local cache.
        @return  @c PDB_OK if fetched and content is a PDB file.
         PDB_NO_CONNECTION if the internet connection could not be established
         to the server.
         PDB_PDBENTRY_NOT_EXISTS if the PDB entry could not be found.
        """
        if( s._CurrentPDBID==pdbID ):
            # There is nothing to do, the file has already been loaded into
            # this object.
            return PDB_OK;
        # This will be set with to the real PDB ID when loaded successfully.
        s._CurrentPDBID = "";
        mirrorPDBFN     = s._GetMirrorPDBFileName(pdbID);
        localPDBFN      = s._GetLocalPDBFileName(pdbID);
        # First, let's see if we have already fetched that file previously.
        if(   os.path.isfile(localPDBFN) ):
            fh = open(localPDBFN);
        # See if there is a local PDB mirror installation. If yes, it must not
        # be complete, so check if the file is there.
        elif( os.path.isfile(mirrorPDBFN) ):
            fh = gzip.GzipFile(mirrorPDBFN);
        # Ok, neither in the local cache, nor in a PDB mirror, get it online.
        else:
            # For speed reasons, we download the PDB file from RCSB as a gzip
            # compressed file and uncompress it to the final local file.
            middle        = pdbID[1:3];
            pdbURL        = s.PDBBaseURL % pdbID;
            try:
                urllib.urlretrieve(pdbURL, localPDBFN);
            except:
                return PDB_NO_CONNECTION;
            fh  = open(localPDBFN);
        s[:]= [line for line in fh];
        fh.close();
        s._GroupLines();
        # After the grouping call, we have split by the record
        # names. Let's see if we have PDB-compliant records, especially
        # those we are interested in: a) hetgroups for ligands, and
        # standard amino acids such that the hetgroups become ligands in a
        # protein chain.
        if( "ATOM  " not in s._Records and "HETATM" not in s._Records ):
            # Obviously this is not the case. This cannot be a PDB file
            # then. It might be an error message (page) from PDB. Wave
            # goodbye. 
            s[:] = [];
            return PDB_PDBENTRY_NOT_EXISTS;
        s._DetermineHetGroups();
        fh = open(localPDBFN, "w");
        fh.write("".join(s));
        fh.close();
        s._CurrentPDBID = pdbID;
        return PDB_OK;

    def FetchMTZ(s, pdbID):
        """Fetch an MTZ file from the EDS server.
        @return  @c PDB_OK or PDB_NO_CONNECTION.
        """
        middle     = pdbID[1:3];
        mtzURL     = s.EDSBaseURL % (middle, pdbID) + pdbID + "_sigmaa.mtz";
        localMTZFN = s._GetLocalMTZFileName(pdbID);
        if( not os.path.isfile(localMTZFN) ):
            try:
                urllib.urlretrieve(mtzURL, localMTZFN);
            except:
                return PDB_NO_CONNECTION;
        # Get the magic string from the file. Should be "MTZ" if it's a real
        # MTZ file. Otherwise it's an error message from EDS.
        fh = open(localMTZFN, "rb");
        mtz3Bytes = fh.read(3);
        fh.close();
        if( mtz3Bytes!="MTZ" ):
            return PDB_EDSENTRY_NOT_EXISTS;
        return PDB_OK;

    def FetchEDSStatLis(s, pdbID):
        """Fetch an '_stat.lis' file from the EDS server.
        @return  PDB_OK, PDB_NO_CONNECTION, or PDB_EDSENTRY_NOT_EXISTS.
        """
        # If we load a new stat.lis file on top of a different PDB structure,
        # we would mess up things. Therefore, invalidate the current PDB ID,
        # if it is not the same.
        if( s._CurrentPDBID!=pdbID ):
            s._CurrentPDBID = pdbID;
        middle    = pdbID[1:3];
        mirrorSLFN= s._GetMirrorEDSStatLisFileName(pdbID);
        localSLFN = s._GetLocalEDSStatLisFileName(pdbID);
        slURL     = s.EDSBaseURL % (middle, pdbID) + pdbID + "_stat.lis";
        if( not os.path.isfile(localSLFN) ):
            if( not os.path.isfile(mirrorSLFN) ):
                # XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
                #return False
                try:
                    urllib.urlretrieve(slURL, localSLFN);
                except:
                    # This is only a connection error. If the file has not been
                    # found on the EDS server, a small web page is going to be
                    # sent with a 404 Not Found title.
                    return PDB_NO_CONNECTION;
            else:
                shutil.copyfile(mirrorSLFN, localSLFN);
        
        eds = CEDSFile();
        # Check if the file is really containing data and not only an error
        # message from the server.
        return PDB_OK if eds.is_valid(localSLFN) else PDB_EDSENTRY_NOT_EXISTS;

    def _PopulateLigandData(s, pdbID):
        """Combine ligand data from the PDB entry and from EDS.
        Call this method here only, if both the PDB file and the Stat list
        file have been fetched before and nothing else in between.
        @return  True if things went ok, False if the pdbID is not the current
         PDB ID that is stored in this object.
        """
        if( s._CurrentPDBID!=pdbID ):
            return False;
        eds   = CEDSFile();
        slFFN = s._GetLocalEDSStatLisFileName(pdbID);
        eds.parse(pdbFFN=slFFN);
        # Data should be ok, since this is an internal function that relies on
        # sanity check of the _stat.lis file before calling this function.
        for rnr, ligand in s._HetGroups.items():
            edsData = eds.get(rnr, None);
            if( edsData!=None ):
                ligand.SetRSCC(edsData[EDSRSCCIDX]);
                ligand.SetOWAB(edsData[EDSOWABIDX]);
        return True;
        
    def FetchPDBnStatLis(s, pdbID):
        """Utility function to get all data at once.
        @return  PDB_OK if everything went ok, files have been loaded from PDB
         and EDS. Otherwise, an error code as defined at the beginning of this
         file will be returned.
        """
        succ = s.FetchPDB(pdbID);
        if( succ!=PDB_OK ):
            return succ;
        succ = s.FetchEDSStatLis(pdbID);
        if( succ!=PDB_OK ):
            return succ;
        succ = s._PopulateLigandData(pdbID);
        if( not succ ):
            return PDB_EDS_INVALID_STATLIS;
        if( len(s._HetGroups)==0 ):
            return PDB_NO_LIGANDS_FOUND;
        return 0;
        
    def WriteCootScript(s, pdbID, resNr, width, height, sigma2FoFc, sigmaFoFc,
                                                        reslist=False):
        # Initially, I started to write guile scripts, but Win$ coot does not
        # have guile, only Python. Converged to Python scripting then, runs on
        # both Linux and Win$.
        # (make-and-draw-map "%s" "FOFCWT" "PHFOFCWT" "" 0 1)
        scriptTemplScm='''(read-pdb "%s")
(auto-read-make-and-draw-maps "%s")
(set-go-to-atom-chain-residue-atom-name "%s" %s "%s")
(set-clipping-front -10.00)
(set-clipping-back -10.00)
(set-graphics-window-size %d %d)
(set-contour-level-in-sigma 1 %3.2f)
(set-contour-level-in-sigma 2 %3.2f)
''';
        pyNavScript = '''
def go2next():
    global cblig
    global badlignum
    cblig = cblig+1 if cblig<badlignum-1 else 0
    set_go_to_atom_chain_residue_atom_name(blchid[cblig], 
                                           blresn[cblig], 
                                           blatom[cblig])
def go2prev():
    global cblig
    global badlignum
    cblig = cblig-1 if cblig>1 else badlignum-1
    set_go_to_atom_chain_residue_atom_name(blchid[cblig], 
                                           blresn[cblig], 
                                           blatom[cblig])
add_key_binding("Next bad ligand", "g", lambda: go2next())
add_key_binding("Previous bad ligand", "v", lambda: go2prev())
cblig=0
blchid, blresn, blatom = [], [], []
'''
        # we need to use raw strings, as the path separator on windows is the
        # escape character...
        scriptTemplPy='''read_pdb(r"%s")
auto_read_make_and_draw_maps(r"%s")
set_go_to_atom_chain_residue_atom_name("%s", %s, "%s")
set_clipping_front(-10.00)
set_clipping_back(-10.00)
set_graphics_window_size(%d, %d)
set_contour_level_in_sigma(1, %3.2f)
set_contour_level_in_sigma(2, %3.2f)
''';

        scriptTempl = scriptTemplPy if s._ScriptingHost=="python"\
                      else scriptTemplScm;
        chnNm  = resNr[0];
        rnr    = resNr[1:5];
        mtzFFN = s._GetLocalMTZFileName(pdbID);
        pdbFFN = s._GetLocalPDBFileName(pdbID);
        scmFFN = s._GetLocalCootScriptFileName(pdbID);
        atomNm = s.GetCenterAtom(resNr);
        script = scriptTempl % (pdbFFN, mtzFFN, chnNm, rnr, atomNm,
                                width, height, sigma2FoFc, sigmaFoFc);
        if reslist and s._ScriptingHost=="python":
            for residue in reslist:
                pyNavScript += 'blchid.append("%s")\n' % residue[0]
                pyNavScript += 'blresn.append(%s)\n' % residue[1:5]
                pyNavScript += 'blatom.append("%s")\n' % s.GetCenterAtom(residue)
            pyNavScript += 'badlignum = len(blchid)'
            script += pyNavScript
        fh = open(scmFFN, "w");
        fh.write(script);
        fh.close();
        return scmFFN;

    def WriteShellScript(s, pdbID):
        cootTmpFiles = "0-coot.state.scm 0-coot.state.py 0-coot-history.scm 0-coot-history.py";
        if( sys.platform=="win32" ):
            # win$ has .bat files and we need to quote everyting due to spaces
            # in path names...
            # do not know why, but we need to delete the temp files first and
            # then start coot. The other way round did not work. *scratch*
            scriptTempl  = "del /Q /F %s" + os.linesep;
            scriptTempl += '"' + s._CootExec + '"' + ' -s "%s"' + os.linesep;
        else:
            # some Unix flavor
            scriptTempl = '#/bin/sh' + os.linesep;
            scriptTempl += 'rm -f %s' + os.linesep;
            scriptTempl += s._CootExec + \
                           ' -s "%s" >/dev/null 2>/dev/null'+os.linesep;
        shFFN  = s._GetLocalShellScriptFileName(pdbID);
        scmFFN = s._GetLocalCootScriptFileName(pdbID);
        script = scriptTempl % (cootTmpFiles, scmFFN);
        #Debug(shFFN);
        fh     = open(shFFN, "w");
        fh.write(script);
        fh.close();
        os.chmod(shFFN, 0755);
        return shFFN;
        


if( __name__=="__main__" ):
    # Large scale testing. Checked all ~50,000 PDB entries with an associated
    # EDS entry.
    from misc import CalcCombinedScore;
    obj = CPDBFile();
    obj.SetPDBDir("/Users/chris/Data/PDB");
    obj.SetEDSDir("./data/EDSDB");
    obj.SetCootTmpDir("/Users/chris/tmp/coot-data");
    print "PDB ID\tLigNm\tRNR\tRSCC\tOWAB\tMolWt\tChrg\tNrHeavy\tNrHyd\tNrAtms\tReleaseDate\tResol\tRwrk\tRFree\tRobs\tSftwre\tJrnl\tPMID\tClass\tScore"
    startFrom = "1n08"; #None;
    for pdbID in open("/tmp/pdbXraylist.tsv"):
        pdbID = pdbID.strip();
        if( startFrom!=None and pdbID!=startFrom ):
            continue ;
        startFrom = None;
        succ=obj.FetchPDBnStatLis(pdbID);
        if( succ==0 ):
            hetGroups = obj.GetAllHetGroups();
            dpl = [rnr for rnr, ligand in hetGroups];
            obj.WriteCootScript(pdbID, rnr, 500, 450, 1.0, 3.0, reslist=dpl);
            obj.WriteShellScript(pdbID)
            for rnr, ligand in hetGroups:
                score = CalcCombinedScore(ligand.GetRSCC(),
                                          obj.GetResolution());
                ligand.SetCombinedScore(score);
                row = [pdbID,
                       ligand.Get3LetterName(),
                       rnr,
                       "%4.3f" % ligand.GetRSCC(),
                       "%4.2f" % ligand.GetOWAB(),
                       "%5.1f" % ligand.GetMolWt(),
                       "%d" % ligand.GetCharge(),
                       "%d" % ligand.GetNrHeavyAtoms(),
                       "%d" % ligand.GetNrHydrogenAtoms(),
                       "%d" % ligand.GetNrAtoms(),
                       obj.GetDepositDate(),
                       "%3.2f" % obj.GetResolution(), 
                       "%4.3f" % obj.GetRWork(), 
                       "%4.3f" % obj.GetRFree(), 
                       "%4.3f" % obj.GetRObs(),
                       obj.GetRefinementSoftware(),
                       obj.GetJournal(),
                       obj.GetPubMedID(),
                       ligand.GetClassification(),
                       "%4.3f" % ligand.GetCombinedScore()];
                print "\t".join(row);
                   
        else:
            if( succ in (1,2,4) ):
                sys.stderr.write("Could not fetch: %s\n" % pdbID);



# eds
# 
# Reads an RDS file and keeps some data (esp. RSCC).
#
# Author       : Chris X. Weichenberger
# Created      : 2010-11-23
# License
# Twilight is a GUI for displaying ligands incompatible with electron density.
#    Copyright (C) 2012  Christian X. Weichenberger, Edwin Pozharski, and
#    Bernhard Rupp.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Modifications:
# YYYY-MM-DD (WHO) what
#
# known bugs:
#
#

import os;
import re;
import sys;
import string;
from misc import EDSDIR, ReadTextFile, STDAMINOS, NONLIGANDS;


EDSRESNMIDX = 0; # residue name
EDSRSCCIDX  = 1; # rscc 
EDSOWABIDX  = 2; # occupany weighted average B factor

class CEDS(dict):
    """Acts like a cache."""
    def __getitem__(s, pdbCode):
        """Key is a PDB four letter code, value is a dictionary (key: rnr,
        value: (resnm, rscc) )"""
        if( pdbCode not in s ):
            edsf = CEDSFile();
            edsf.parse(pdbCode);
            s[pdbCode] = {}
            if( edsf.has_features() ):
                for rnr, value in edsf.items():
                    s[pdbCode][rnr] = value;
        return dict.__getitem__(s, pdbCode);


class CEDSFile(dict):
    def __init__(s):
        dict.__init__(s);
        s.FileFound = False;
    
    def __str__(s):
        # a key is a residue number with chain name.
        keys = s.keys();
        keys.sort();
        lst  = ["%s,%s\t%s\t%s\t%4.3f" % \
                  (s.PDBCode,k[0], k, s[k][EDSRESNMIDX], s[k][EDSRSCCIDX])\
                  for k in keys];
        return string.join(lst, "\n");
    
    def parse(s, pdbCode="", pdbFFN=""):
        """Read an EDS file.
        """
        if( pdbCode!="" ):
            s.PDBCode = pdbCode;
            ffn = os.path.join(EDSDIR, pdbCode + "_stat.lis");
        else:
            s.PDBCode = "????";
            ffn = pdbFFN;
        lines = ReadTextFile(ffn);
        if( lines!=None and len(lines)!=0 ):
            s.FileFound = True;
            for line in lines:
                if( line[0]!="!" ):
                    resNm = line[8:11].strip();
                    if( resNm in NONLIGANDS ): continue ;
                    rnr   = line[12:18];
                    owab  = float(line[34:39]);
                    chnNm = rnr[0];
                    rscc  = float(line[19:26]);
                    s[rnr]= (resNm, rscc, owab);

    def has_features(s):
        return s.FileFound;

    def is_valid(s, pdbFFN):
        """Implements some simple heuristics whether this is a valid EDS Stat
        List file. 
        @return @c True or @c False.
        """
        # There should be a line like
        # ! Created by MAPMAN V. 080625/7.8.5 at Mon Jun 8 22:36:34 2009 [...]
        # We will look for this in the first couple of lines.
        readAhead = 10;
        fh        = open(pdbFFN);
        i         = 1;
        for line in fh:
            if( re.search("^!.*MAPMAN", line) ):
                return True;
            i += 1;
            if( i>readAhead ):
                break ;
        return False;
            
        
def Usage():
    print """Usage %s -f pdbcode
 where  -f Parses and prints an EDS file.""" % \
    (os.path.basename(sys.argv[0]), );
    
if( __name__=="__main__" ):
    if( len(sys.argv) not in (2,3) ):
        Usage();
        sys.exit(1);
    if( sys.argv[1] not in ("-f", ) ):
        Usage();
        sys.exit(1);
        
    if( sys.argv[1]=="-f" ):
        edsf = CEDSFile();
        edsf.parse(sys.argv[2]);
        if( edsf.has_features() ):
            print edsf



# Periodic table of elements.
#
# Author       : Chris X. Weichenberger
# Created      : 2012-08-02
# License
# Twilight is a GUI for displaying ligands incompatible with electron density.
#    Copyright (C) 2012  Christian X. Weichenberger, Edwin Pozharski, and
#    Bernhard Rupp.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Modifications:
# YYYY-MM-DD (WHO)  what

from misc import ELEMENTSFN, CTable;

class CElement:
    def __init__(s, Z, symbol, element, weight):
        s._Z       = int(Z);
        s._Symbol  = symbol;
        s._Element = element;
        s._Weight  = float(weight);

    def __str__(s):
        return "%d %s (%s) %7.5f" % (s._Z, s._Symbol, s._Element, s._Weight);
    
    def GetZ(s):
        return s._Z;

    def GetSymbol(s):
        return s._Symbol;

    def GetElement(s):
        return s._Element;

    def GetWeight(s):
        return s._Weight;


class CPeriodicTable(dict):
    """Periodic table of the elements.
    This is a Dictionary with the element symbol as key and a CElement
    instance as its value.
    """
    def __init__(s):
        tab = CTable(ELEMENTSFN, ints={"Z": "%d"}, floats={"weight": "%7.5f"});
        for line in tab.AsTable():
            z, sym, elem, wt = line;
            s[sym.upper()] = CElement(z, sym, elem, wt);

__PERIODIC_TABLE = CPeriodicTable();

def GetPeriodicTable():
    """Get the global CPeriodicTable object (performance related)
    """
    return __PERIODIC_TABLE;

if( __name__=="__main__" ):
    for sym, elem in GetPeriodicTable().items():
        print elem



# Ligand data container.
#
# Author       : Chris X. Weichenberger
# Created      : 2012-07-26
# License
# Twilight is a GUI for displaying ligands incompatible with electron density.
#    Copyright (C) 2012  Christian X. Weichenberger, Edwin Pozharski, and
#    Bernhard Rupp.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Modifications:
#  2012-08-02 (CW)  Computation of ligand molecular weight.

NOTHING           = "-";
ISBUFFERMOLECULE  = "B";
ISLPEPTIDE        = "P";

import sys

from misc import ReadTextFile, LPETIDESFN, BUFFERNMFN, NOVALIDFLOATSTR;
from chemform import CChemicalFormulaData;
from elements import GetPeriodicTable;


__PEPLIST = ReadTextFile(LPETIDESFN);
__BUFLIST = ReadTextFile(BUFFERNMFN);

def IsPeptide(nm):
    if( nm in __PEPLIST ):
        return ISLPEPTIDE;
    return NOTHING;


def IsBufferMolecule(nm):
    if( nm in __BUFLIST ):
        return ISBUFFERMOLECULE;
    return NOTHING;


def clsOR(c1, c2):
    """Performs logical OR on two classification codes."""
    if( c1==NOTHING and c2==NOTHING ):
        return NOTHING;
    elif( c1==NOTHING ):
        return c2;
    elif( c2==NOTHING ):
        return c1;
    # now, neither of them is NOTHING.
    elif( c1 in (ISION,ISCHARGEDMOLECULE) ):
        return c2;
    elif( c2 in (ISION,ISCHARGEDMOLECULE) ):
        return c1;
    else:
        sys.stderr.write("ligand.py: Uuups. Total confusion. Please report.");
        # We should exit here.
        return c1;


class CLigand:
    def __init__(s):
        s._LigName = "N/A";
        s._ResNr   = "N/A";
        s._RSCC    = float(NOVALIDFLOATSTR);
        s._OWAB    = float(NOVALIDFLOATSTR);
        s._MolWt   = float(NOVALIDFLOATSTR);
        s._ChemForm = CChemicalFormulaData("");
        
    def __str__(s):
        return "%s(%s) RSCC: %s OWAB: %s" %\
               (s._ResNr, s._LigName, s._RSCC, s._OWAB);

    def Set3LetterName(s, val):
        s._LigName = val;
        # Assign the classification code.
        s._ClsCode = clsOR(NOTHING, IsPeptide(s._LigName));
        s._ClsCode = clsOR(s._ClsCode, IsBufferMolecule(s._LigName));
    
    def SetResNr(s, val):
        s._ResNr = val;
    
    def SetRSCC(s, val):
        s._RSCC = float(val);

    def SetOWAB(s, val):
        s._OWAB = float(val);

    def SetMolWt(s, val):
        s._MolWt = float(val);
    
    def SetCombinedScore(s, val):
        s._Score = float(val);
    
    def SetChemFormula(s, formula):
        s._ChemForm = CChemicalFormulaData(formula);

    def ComputeMolWeight(s):
        """Compute the molecular mass of this ligand. (More or less.)
        If an unknown chemical element is encountered, the reported molecular
        weight is set to @c NOVALIDFLOATSTR.
        @return Molecular mass suitable for SetMoWt().
        """
        weight = 0.0;
        for sym, nr in s._ChemForm:
            elem = GetPeriodicTable().get(sym, None);
            if( elem!=None ):
                weight += nr * elem.GetWeight();
            else:
                sys.stderr.write("Unknown chemical element: ``%s''.\n" % sym);
                return float(NOVALIDFLOATSTR);
        return weight;
    
    def Get3LetterName(s):
        """@return @c String """
        return s._LigName;
    
    def GetResNr(s):
        """@return @c String """
        return s._ResNr;
    
    def GetRSCC(s):
        """@return @c Float """
        return s._RSCC;

    def GetOWAB(s):
        """@return @c Float """
        return s._OWAB;

    def GetMolWt(s):
        """@return @c Float """
        return s._MolWt;
    
    def GetCombinedScore(s):
        return s._Score;
    
    def GetClassification(s):
        """@return @c String """
        return s._ClsCode;
    
    def GetCharge(s):
        """@return @c Int """
        return s._ChemForm.getCharge();

    def GetNrHeavyAtoms(s):
        """@return @c Int """
        return s._ChemForm.nrHeavyAtoms();
        
    def GetNrHydrogenAtoms(s):
        """@return @c Int """
        return s._ChemForm.nrHydrogenAtoms();

    def GetNrAtoms(s):
        """@return @c Int """
        return s._ChemForm.nrAtoms();
    
if( __name__=="__main__" ):
    lig = CLigand();
    lig.Set3LetterName("IPA");
    lig.SetChemFormula("  2(C3 H8 O)    ");
    print lig.ComputeMolWeight();



# Some basic file io.
#
# Author       : Chris X. Weichenberger
# Created      : 2012-02-24
# License
# Twilight is a GUI for displaying ligands incompatible with electron density.
#    Copyright (C) 2012  Christian X. Weichenberger, Edwin Pozharski, and
#    Bernhard Rupp.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Modifications:
# YYYY-MM-DD (WHO) what
#

import os
import sys

def Output(text):
    """Write text + newline to stderr."""
    sys.stderr.write(text + os.linesep);

def Debug(thing):
    fh = open("/tmp/debug-dysplayer.txt", "a");
    fh.write(str(thing) + os.linesep);
    fh.close();

def GetVersionedInOutFileNames(fn):
    """Get the next versioned file name.
    - fn: Current file name. May be versionless or versioned.
    * return: 2-Tuple: (last existing versioned file name, next versioned file
      name which does not yet exist)
    """
    base, ext = os.path.splitext(fn);
    version   = 1;
    if( base[-5:-3]=="-V" ):
        try:
            version = int(base[-3:]);
            base    = base[:-5];
        except ValueError:
            # fall back to version number one.
            pass ;
    usedFFN = fn;
    # quite ugly, but save...
    for verNr in range(999, version-1, -1):
        ffn = base + "-V%03d" % verNr + ext;
        if( not os.path.exists(ffn) ):
            unusedFFN = ffn;
            continue ;
        usedFFN = ffn;
        break ;
    return usedFFN, unusedFFN;
